package org.embl.gbcs.jemultiplexer;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import net.sf.picard.cmdline.CommandLineProgram;
import net.sf.picard.cmdline.Option;
import net.sf.picard.cmdline.StandardOptionDefinitions;
import net.sf.picard.cmdline.Usage;
import net.sf.picard.fastq.FastqReader;
import net.sf.picard.fastq.FastqRecord;
import net.sf.picard.fastq.FastqWriter;
import net.sf.picard.sam.FastqToSam;
import net.sf.picard.util.FastqQualityFormat;
import net.sf.picard.util.SolexaQualityConverter;
import net.sf.picard.util.QualityEncodingDetector;
import net.sf.samtools.SAMUtils;

import org.apache.log4j.Logger;
import org.embl.cg.utilitytools.utils.ExceptionUtil;
import org.embl.cg.utilitytools.utils.FileUtil;
import org.embl.cg.utilitytools.utils.StringUtil;
import org.embl.cg.utilitytools.utils.parser.csv.CSVLine;
import org.embl.cg.utilitytools.utils.parser.csv.InvalidHeaderException;
import org.embl.gbcs.embase.api.EmBaseDatabaseFacade;
import org.embl.gbcs.embase.api.Queries;
import org.embl.gbcs.embase.api.exception.EmBASEConnectionException;
import org.embl.gbcs.embase.api.model.Item;
import org.embl.gbcs.embase.api.model.NGSLibrary;

import com.developpez.adiguba.shell.ProcessConsumer;




public class Jemultiplexer extends CommandLineProgram {
	private static Logger log = Logger.getLogger(Jemultiplexer.class);

	/**
	 * Defaults
	 */
	protected static final BarcodePosition DEFAULT_BARCODE_READ_POS = BarcodePosition.BOTH;
	protected static final BarcodePosition DEFAULT_BARCODE_FOR_SAMPLE_MATCHING = BarcodePosition.BOTH;
	protected static final boolean DEFAULT_REDUNDANT_BARCODES = true;
	protected static final boolean DEFAULT_STRICT = false;
	protected static final Integer DEFAULT_MAX_MISMATCHES = 1;
	protected static final Integer DEFAULT_MIN_MISMATCH_DELTA = 1;
	protected static final Integer DEFAULT_MIN_BASE_QUALITY = 10;
	protected static final Integer DEFAULT_XTRIMLEN = 1;
	protected static final Integer DEFAULT_ZTRIMLEN = 0;
	protected static final boolean DEFAULT_CLIP_BARCODE = true;
	protected static final boolean DEFAULT_ADD_BARCODE_TO_HEADER = true;
	protected static final FastqQualityFormat DEFAULT_QUALITY_FORMAT = FastqQualityFormat.Standard; 
	protected static final boolean DEFAULT_GZIP_OUTPUTS = true;
	protected static final boolean DEFAULT_KEEP_UNASSIGNED_READ = true;
	protected static final boolean DEFAULT_WRITER_FACTORY_USE_ASYNC_IO = true;
	protected static final String DEFAULT_UNASSIGNED_FILE_NAME_2 = "unassigned_2.txt";
	protected static final String DEFAULT_UNASSIGNED_FILE_NAME_1 = "unassigned_1.txt";
	protected static final String DEFAULT_METRICS_FILE_NAME = "jemultiplexer_out_stats.txt";
	
	
	
	
	/**
	 * test mode : tell the code to stop as soon as the doWork() is called i.e. allows to test the parsing in order to check 
	 * the status after parsing.
	 */
	protected boolean TEST_MODE_STOP_AFTER_PARSING = false;
	
	@Usage(programVersion="1.0.4") 
	public String USAGE = "\tJemultiplexer : A fastq files demultiplexer with many neat options.\n"
			+ "\tInput fastq files can be in gzip compressed format (end in .gz). \n" 
			+ "\tBy default output files are gzipped and have names following the pattern \n"
			+"\t'<samplename>_<barcode[-barcode2...-barcodeN]>_[1|2].txt[.gz]' unless you gave file\n" +
			"\t names to use within the barcode description file.\n" 
			+"\tSee 'Usage' section below for all other option defaults\n"+
			"Usage: \n"
			+"\tjava -Xmx2g -jar jemultiplexer.jar F1=fastq_1.txt.gz F2=fastq_2.txt.gz BF=barcodes.bs O=/Users/girardot/Desktop/test-jemultiplexer2/ BPOS=BOTH MM=1 MMD=1 Q=10 XT=1 ZT=0 DIAG=diags.txt UN=true GZ=true CREATE_MD5_FILE=true & \n" 
			;


	@Option(shortName="F1", optional = false,
			doc="Input fastq file (optionally gzipped) for single end data, or first read in paired end data.\n"
			)
	public File FASTQ_FILE1;

	@Option(shortName="F2", optional = true,
			doc="Input fastq file (optionally gzipped) for the second read of paired end data.\n"
			)
	public File FASTQ_FILE2 = null;

	@Option(shortName="I1", optional = true,
			doc="Fastq file for index 1 (optionally gzipped) i.e. when barcodes are read with specific primers (the Illumina way basically).\n" +
					"When using INDEX_FILE, **sample encoding** barcodes\n"+
					" cannot be found within the reads anymore (BM option set to NONE automatically, see below).\n"+
					"\tNevertheless, barcodes (for e.g. molecular barcoding or for any purpose other than sample encoding) \n"+
					"can be still found in the read(s) and the BPOS option is then used to indicate the presence of such barcode(s).\n"+
					"\tWhile the options Q, MM, MMD, S, BRED apply to the INDEX_FILE barcodes; \n"+
					"the options XT, ZT, CLIP will always apply to barcodes found in reads only.\n"
			)
	public File INDEX_FILE1;

	@Option(shortName="I2", optional = true,
			doc="Fastq file for index 2 (optionally gzipped) i.e. when barcodes are read with specific primers (the Illumina way basically).\n" +
					"A INDEX_FILE1 MUST be provided when INDEX_FILE2 is given. This situation corresponds to using 2 extra specific primers\n"+
					" for index sequencing (not clear whether this even exists right now!).\n"+
					" When using INDEX_FILE, **sample encoding** barcodes\n"+
					" cannot be found within the reads anymore (BM option set to NONE automatically, see below).\n"+
					"\tNevertheless, barcodes (for e.g. molecular barcoding or for any purpose other than sample encoding) \n"+
					"can be still found in the read(s) and the BPOS option is then used to indicate the presence of such barcode(s).\n"+
					"\tWhile the options Q, MM, MMD, S, BRED apply to the INDEX_FILE barcodes; \n"+
					"the options XT, ZT, CLIP will always apply to barcodes found in reads only.\n"
			)
	public File INDEX_FILE2 = null;

	/*
	 * A boolean to indicate that sample barcoding comes from separate index file(s)
	 */
	protected boolean USE_SAMPLE_INDEX_FILES = false;
	
	
	/**
	 * Are we running SE or PE ? Option set at cmd line parsing time
	 */
	protected boolean RUNNING_PAIRED_END = false;

	
	@Option(shortName="BF", optional = false,  mutex={"USE_EMBASE"},
			doc="Barcode file describing sequence list and sample names. "
					+"Tab-delimited file with 2 columns, with the sample in col1 and the corresponding barcode in col2.\n" +
					"Simple barcode file format : 2 tab-delimited colums\n"+
					"\t"+"If multiple barcode map to the same sample, either line can be duplicated e.g.\n" +
					"\t"+"\t"+"sample1\tATAT\n" +
					"\t"+"\t"+"sample1\tGAGG\n" +
					"\t"+"\t"+"sample2\tCCAA\n" +
					"\t"+"\t"+"sample2\tTGTG\n" +
					"\t"+"Or barcodes can be combined using the OR operator '|' i.e. the file above can be re-written like\n " +
					"\t"+"\t"+"sample1\tATAT|GAGG\n" +
					"\t"+"\t"+"sample2\tCCAA|TGTG\n" +
					"\t"+"Finally, for the special situation of paired-end data in which barcodes differ at both ends " +
					"(ie BPOS=BOTH BRED=false BM=BOTH , see BRED option description), barcodes for read_1 and read_2 can be" +
					" distinguished using a ':' separator i.e. \n" +
					"\t"+"\t"+"sample1\tATAT:GAGG\n" +
					"\t"+"\t"+"sample2\tCCAA:TGTG\n" +
					"\t"+"Here understand that sample 1 is encoded with ATAT barcode at read_1 AND GAGG barcode at read_2. " +
					"Note that you can still combine barcodes using | e.g. \n"+
					"\t"+"\t"+"sample1\tATAT|GAGG:CCAA|TGTG\n" +
					"\t"+"would mean that sample 1 is mapped by the combination of barcode: ATAT OR GAGG at read_1 AND CCAA OR TGTG at read_2.\n"+
					"Extended barcode file format : 3 (single-end) or 4 (paired-end) tab-delimited colums\n"+
					"\t"+"same as the simple barcode file format but the extra columns contains the file name(s) to use to name output files :" +
					" only one extra column for single-end and 2 extra columns for paired-end. In case, lines are duplicated (multiple barcodes" +
					"mapping the same sample), the same file name should be indicated in the third (and fourth) column(s). \n"+
					"\t"+"\t"+"sample1\tATAT\tspl1_1.txt.gz\tspl1_2.txt.gz\n" +
					"\t"+"\t"+"sample1\tGAGG\tspl1_1.txt.gz\tspl1_2.txt.gz\n" +
					"\t"+"\t"+"sample2\tCCAA\tspl2_1.txt.gz\tspl2_2.txt.gz\n" +
					"\t"+"Or\n" +
					"\t"+"\t"+"sample1 \t ATAT|GAGG:CCAA|TGTG \t spl1_1.txt.gz \t spl1_2.txt.gz\n" +
					"Ns in barcode sequence are allowed and are used to flag positions that should be ignored in sample matching \n"+
					"\t i.e. they will be clipped off the read sequence (like in iCLIP protocol). If you need to get access to the \n" +
					"\t nature of these positions in downstream processing steps (e.g. PCR dups elimination), consider enabling the METRICS_FILE_NAME option."
			)
	public File BARCODE_FILE = null;
	
	@Option(shortName="EM", optional = false, mutex={"BARCODE_FILE"},
			doc="Tells Jemultiplexer to fetch information from emBASE and place demultiplexed files directly in emBASE repository structure.\n" +
					"This option is mutually exclusive with BARCODE_FILE.\n" +
					"This option cannot be used when using INDEX_FILE.\n" +
					"Note that using this option forces O=null GZ=true UN=true UF1=null UF2=null STATS_ONLY=false (all other user options supported).\n" 
			)
	public boolean USE_EMBASE = false;
	
	
	
	/**
	 * object in charge of validating the barcode file and holds useful maps ; to be set at cmd line validation
	 */
	protected BarcodeValidator barcodeValidator = null; 
	
	@Option(shortName="BPOS", optional = true,
			doc="For paired-end data, where to expect the barcode(s) : READ_1 (beginning of read from FASTQ_FILE_1),"
					+" READ_2 (beginning of read from FASTQ_FILE_2), BOTH (beginning of both reads).\n"+
					"Automatically set to READ_1 in single end mode.\n"+
					"If INDEX_FILE otions is/are used, BPOS is then used to indicate the presence of additional barcode(s) in the read(s).\n"+
					"Note that by default, BPOS=NONE is assumed when INDEX_FILE option(s) is/are used."+
					"\t!! Importantly, these additional barcodes must not encode sample identity information (BM is forced to NONE) but used for \n"+
					"\te.g. molecular barcoding or for any purpose other than sample encoding."
			)
	public BarcodePosition BARCODE_READ_POS = DEFAULT_BARCODE_READ_POS ;

	@Option(shortName="BRED", optional = true,
			doc="For paired-end data with BARCODE_READ_POS == BOTH, or both INDEX_FILE1 and INDEX_FILE2 are provided ;\n" +
					" this option indicates if both read's barcodes encode redundant information, which is the usual situation (REDUNDANT_BARCODES=true)\n" +
					" i.e. barcodes are supposed to be the same at both ends (or in both INDEX_FILE) or to resolve to the\n" +
					" same sample (when a pool of barcodes has been used for each sample).\n "+
					"\tWhen REDUNDANT_BARCODES=false (and INDEX_FILE are NOT used, see next section when using two INDEX_FILE), the 2 barcodes potentially encode\n" +
					" different information. For example, only one of the barcodes encodes the sample the read belongs to while \n" +
					"the second barcode might be a random barcode to tell apart PCR artefacts from real duplicates.\n" +
					"Another example is when both barcodes should be used in a combined fashion to resolve the sample.\n" +
					"In the first example, you should use BPOS=BOTH BRED=false BM=READ_1.\n" +
					"In the second example, you should have BPOS=BOTH BRED=false BM=BOTH. Note that with BPOS=BOTH BRED=true BM=BOTH,\n" +
					" the behavior would be different as Jemultiplexer would then check the STRICT option to perform sample resolution.\n" +
					"Importantly, when BARCODE_READ_POS == BOTH AND REDUNDANT_BARCODES=false, BLEN, barcode matching options" +
					" (MM, MMD, Q) and read trimming/clipping options (XT, ZT) accept different values for both barcodes in " +
					"the form X:Z where X and Z are 2 integers.\n"+
					"\tWhen using 2 INDEX_FILE and REDUNDANT_BARCODES=false, BPOS and BM are ignored (indeed, BPOS is used to indicate \n" +
					"the presence of extra barcoding sequences in reads, and BM=NONE), either BRED=true and the S option will guide the sample\n" +
					" lookup behavior or BRED=false meaning that both barcode should be combined prior to sample lookup.\n" +
					"When BRED=false, barcode matching options (MM, MMD, Q) accept different values for both barcodes in \n" +
					"the form X:Z where X and Z are 2 integers (like when not using INDEX_FILE). BUT BLEN, read trimming/clipping\n" +
					" options (XT, ZT) then apply to extra barcodes (BPOS!=NONE) if present (or are irrelevant when BPOS=NONE)."
			)
	public boolean REDUNDANT_BARCODES = DEFAULT_REDUNDANT_BARCODES;


	@Option(shortName="BM", optional = true,
			doc="Indicates which barcode(s) should be used for sample lookup -- this option is irrelevant when using INDEX_FILE option(s).\n" +
					"Automatically set to READ_1 in single end mode. \n" +
					"For paired-end data and when BARCODE_READ_POS == BOTH, which barcode should be used to resolve sample :\n"
					+"\t"+"- use BM=READ_1 (beginning of read from FASTQ_FILE_1) if only this read should be used for sample matching,\n"
					+"\t"+"- use BM=READ_2 (beginning of read from FASTQ_FILE_2) if only this read should be used for sample matching,\n"
					+"\t"+"- use BM=BOTH (beginning of both reads) if both should be used ; when BM=BOTH, the behaviour of Jemultiplexer is " +
					"different based on the value of REDUNDANT_BARCODES.\n"
					+"\t"+"\t"+"If REDUNDANT_BARCODES=true, the two barcodes are considered to map to the same sample and Jemultiplexer" +
					" uses the two barcodes according to the STRICT value.\n" +
					"\t"+"\t"+"If REDUNDANT_BARCODES=false, the barcode file should map a couple of barcode to each sample (e.g. sample1 => AGAGTG:TTGATA)" +
					" and Jemultiplexer needs both barcodes to find the relevant sample. Note that this is the only situation in which all barcode matching options" +
					" (MM, MMD, Q) accept different values for both barcodes in the form X:Z where X and Z are 2 integers.\n"+
					"!! Important !! . "
			)
	//turned to NONE WHEN INDEX FILES ARE USED
	public BarcodePosition BARCODE_FOR_SAMPLE_MATCHING = BarcodePosition.BOTH;


	@Option(shortName="S", optional = true,
			doc="For paired-end data and when BARCODE_READ_POS == BOTH and BM=BOTH and BRED=true (or when using 2 INDEX_FILE with BRED=true),\n" +
					" tells whether both barcodes should resolve to the same sample.\n"
					+" When true and if only one of the two reads has a barcode match, the read pair is ignored.\n"
					+" When false and if only one of the two reads has a barcode match, the read pair is assigned to the\n"
					+" corresponding sample ; in cases where reads resolve to different samples, the read pair is ignored.\n"
			)
	public boolean STRICT = false;

	@Option(shortName = "LEN", optional = true,
			doc = "Length of the barcode sequences, optional. Taken from barcode file when not given.\n" +
					"In situations where BARCODE_READ_POS == BOTH AND REDUNDANT_BARCODES=false, two distinct" +
					" length can be provided using the syntax LEN=X:Z where X and Z are 2 integers representing" +
					" the barcode length for read_1 and read_2 respectively.\n"+
					"IMPORTANT : when using INDEX_FILE, this MUST be given if BPOS!=NONE as BLEN then indicates the length of the non sample encoding barcodes found in reads.\n" +
					"            If BPOS=NONE, this option is otherwise totally ignored when using sample index files"
			)
	public String BCLEN=null ; //BCLEN parsed to set bclen1 and bclen2
	protected Integer BCLEN_1 = 0;
	protected Integer BCLEN_2 = 0;


	//barcode matching options

	@Option(shortName = "MM", optional = true,
			doc="Maximum mismatches for a barcode to be considered a match. MM=null is like MM=0\n" +
					"In situations where both barcodes are used for sample matching i.e. BPOS=BOTH BM=BOTH (or 2 INDEX_FILE given)\n" +
					" (note that most likely BRED=false as it does not make great sense otherwise), two distinct\n" +
					" values can be given here using the syntax MM=X:Z where X and Z are 2 integers to use for read_1 and read_2 respectively.\n"
			)
	public String MAX_MISMATCHES = DEFAULT_MAX_MISMATCHES.toString();
	public Integer MAX_MISMATCHES_1 = DEFAULT_MAX_MISMATCHES;
	public Integer MAX_MISMATCHES_2 = DEFAULT_MAX_MISMATCHES;

	@Option(shortName = "MMD", optional = true,
			doc="Minimum difference between number of mismatches in the best and second best barcodes for a barcode to be considered a match. MMD=null is like MMD=0\n" +
					"In situations where both barcodes are used for sample matching i.e. BPOS=BOTH BM=BOTH (or 2 INDEX_FILE given)\n" +
					" (note that most likely BRED=false as it does not make great sense otherwise), two distinct\n" +
					" values can be given here using the syntax MMD=X:Z where X and Z are 2 integers to use for read_1 and read_2 respectively.\n"
			)
	public String MIN_MISMATCH_DELTA = DEFAULT_MIN_MISMATCH_DELTA.toString();
	public Integer MIN_MISMATCH_DELTA_1 = DEFAULT_MIN_MISMATCH_DELTA;
	public Integer MIN_MISMATCH_DELTA_2 = DEFAULT_MIN_MISMATCH_DELTA;


	@Option(shortName="Q", optional = true,
			doc="Minimum base quality. Any barcode bases falling below this quality will be considered a mismatch even in the bases match. Q=null is like Q=0.\n" +
					"In situations where both barcodes are used for sample matching i.e. BPOS=BOTH BM=BOTH (or 2 INDEX_FILE given)\n" +
					" (note that most likely BRED=false as it does not make great sense otherwise), two distinct\n" +
					" values can be given here using the syntax Q=X:Z where X and Z are 2 integers to use for read_1 and read_2 respectively.\n"
			)
	public String MIN_BASE_QUALITY = DEFAULT_MIN_BASE_QUALITY.toString();
	public Integer MIN_BASE_QUALITY_1 = DEFAULT_MIN_BASE_QUALITY;
	public Integer MIN_BASE_QUALITY_2 = DEFAULT_MIN_BASE_QUALITY;


	@Option(shortName = "XT", optional = true,
			doc = "Extra number of base to be trimmed right after the barcode (only used if CLIP_BARCODE=true). " +
					"Default is 1 as an extra 'T' (or 'A' depending how you see it) is added for barcode ligation but this default will be " +
					"adapted according to the rules below unless using INDEX_FILE with BPOS=NONE, in which case XT=0.\n" +
					"When running paired-end, two distinct values can be given using the syntax XT=X:Z where X and Z are 2 integers" +
					" to use for read_1 and read_2 respectively. Note that even when BPOS=READ_1 or BPOS=READ_2, a X:Y synthax can be given to " +
					"trim the read w/o barcode as to end up with reads of the same length (note that this can also be operated using ZT). " +
					"If a unique value is given, e.g. XT=1, while running paired-end the following rule applies : " +
					"(1) BPOS=READ_1 or BPOS=READ_2, no trim is applied at the read w/o barcode ; (2) BPOS=BOTH, the value is used for both reads.\n"+
					"Note that XT=null is like XT=0.\n"
			)
	public String XTRIMLEN= DEFAULT_XTRIMLEN.toString() ; //must be further inspected to set XTRIMLEN_1 & XTRIMLEN_2
	public Integer XTRIMLEN_1=null ; // we do not set defaults as having null value here is part of the reasoning !!
	public Integer XTRIMLEN_2=null ; // we do not set defaults as having null value here is part of the reasoning !!

	@Option(shortName = "ZT", optional = true,
			doc = "Extra number of bases to be trimmed from the read end i.e. 3' end. Pretty handy when a pipeline is set and you already" +
					" know you'll trim read at a given size. ZT=null is like ZT=0.\n" +
					"When running paired-end, two distinct values can be given here using the syntax ZT=X:Z where X and Z are 2 integers" +
					" to use for read_1 and read_2 respectively. Note that even when BPOS=READ_1 or BPOS=READ_2, a X:Y synthax can be given to " +
					"trim the read w/o barcode as to end up with reads of the same length (note that this can also be operated using XT). " +
					"Note that if a single value is passed, the value always apply to both reads in paired-end mode without further consideration.\n"
			)
	public String ZTRIMLEN=DEFAULT_ZTRIMLEN.toString() ;
	public Integer ZTRIMLEN_1=DEFAULT_ZTRIMLEN ;
	public Integer ZTRIMLEN_2=DEFAULT_ZTRIMLEN ;

	@Option(shortName="C", optional = true,
			doc="Remove barcode sequence from read, as well as XTRIMLEN (and ZTRIMLEN) bases if applicable, before writing to output file.\n"
					+" If false, reads are written without modification to output file. \n" +
					"Apply to both barcodes when BPOS=BOTH.\n"+
					"When using INDEX_FILE, only applies to barcodes found in read(s).\n"
			)
	public boolean CLIP_BARCODE = DEFAULT_CLIP_BARCODE;

	@Option(shortName="ADD", optional = true,
			doc="Add matched barcode at the end of the read header. Apply to both barcodes when BPOS=BOTH.\n"
					+"\t"+"If true, the string ':barcode' is added at the end of the read header with " 
					+"a ':' added only if current read header does not end with ':'.\n"
					+"\t"+"If both reads of the pair have a barcode (i.e. BARCODE_READ_POS == BOTH), then" 
					+"the second read also has its own matched barcode written. Else, the read without a barcode "
					+"receives the barcode from the barcoded read.\n"
					+"\t"+"For example :\n"
					+"\t"+"\t" +"'@D3FCO8P1:178:C1WLBACXX:7:1101:1836:1965 2:N:0:'\n"
					+"\t" +"becomes\n"
					+"\t"+"\t" +"'@D3FCO8P1:178:C1WLBACXX:7:1101:1836:1965 2:N:0:BARCODE'\n"
					+"\n\tIn the very special situation of barcodes containing random positions, i.e. 'N', (for example like " +
					"\tin the iCLIP protocol), the added sequence is the clipped sequence and NOT the matched barcode.\n"+
					"When using INDEX_FILE(s), the rule is the following:\n"+
					"\t"+"First the sample encoding barcodes from I1 (and I2 when relevant) are added to the barcode I1:I2 like \n"+
					"\t"+"\t"+"'@D3FCO8P1:178:C1WLBACXX:7:1101:1836:1965 2:N:0:I1_BARCODE:I2_BARCODE'\n"+
					"\t"+"Then, if BPOS!=NONE, the sequence clipped from the read(s) are added to their own header, like \n"+
					"\t"+"\t"+"'@D3FCO8P1:178:C1WLBACXX:7:1101:1836:1965 2:N:0:I1_BARCODE:I2_BARCODE:CLIPPED_SEQ_FROMREAD'\n"
			)
	public boolean ADD_BARCODE_TO_HEADER  = DEFAULT_ADD_BARCODE_TO_HEADER;

	@Option(shortName = "RCHAR", optional = true,
			doc="Replace spaces in read name/header using provided character. This is particularly handy when you need to retain" 
					+"\t ADDed barcode in read name/header during mapping (everything after space in read name is usually clipped in BAM files)." 
					+"\t"+"For example, with RCHAR=':' :\n"
					+"\t"+"\t" +"'@D3FCO8P1:178:C1WLBACXX:7:1101:1836:1965 2:N:0:'\n"
					+"\t" +"becomes\n"
					+"\t"+"\t" +"'@D3FCO8P1:178:C1WLBACXX:7:1101:1836:1965:2:N:0:BARCODE'\n"
			)
	public String READ_NAME_REPLACE_CHAR = null;
	
	
	//UNCLEAR IF EVER USEFUL !! WAIT BEFORE IMPLEMENTING
//	@Option(shortName = "RED", optional = true,
//			doc="Removes all positions flagged as N in clipped sequences before adding them to the header.This only makes sense if ADD=true and barcode sequences contain N positions.\n"+
//					"For example, if you indicated a barcode like NNNATGANN and the sequence extracted from the read is GGGATGATT :\n"+
//					"\t"+"RED=false => GGGATGATT is written in fastq header\n"+
//					"\t"+"RED=false => ATGA is written in fastq header\n"
//			)
//	public boolean REDUCE_BARCODE_N = false;
	

	/*
	 * More general options below
	 */

	@Option(shortName="V", optional = true,
			doc="A value describing how the quality values are encoded in the fastq.  Either 'Solexa' for pre-pipeline 1.3 " +
					"style scores (solexa scaling + 66), 'Illumina' for pipeline 1.3 and above (phred scaling + 64) or 'Standard' for phred scaled " +
					"scores with a character shift of 33.  If this value is not specified (or 'null' is given), the quality format will be detected automatically.\n"
			)
	public FastqQualityFormat QUALITY_FORMAT = DEFAULT_QUALITY_FORMAT;

	@Option(shortName = StandardOptionDefinitions.METRICS_FILE_SHORT_NAME, optional = true,
			doc="File name where to write demultiplexing statistics. "
					+"Either a name (in which case the file will be created in the output dir) or full path.\n"
			)
	public String METRICS_FILE_NAME = DEFAULT_METRICS_FILE_NAME;
	File metricsFile = null; //final File object to use with METRICS_FILE_NAME, set in cmdline validation 

	/**
	 * This is a tab-delimited 7 columns file:
	 * - Read_1_Header and Read_2_Header : fastq headers of read 1 (fwd read ie from fastq file 1) and 2 (rev read ie from fastq file 2)
	 * - seq1 and seq2 : seq as extracted from the reads 1 and 2 
	 * - bc1 and bc2 : barcode after remapping ; null if no remapping could be made
	 * - sample : final decision ie what sample was used for this read pair
	 */
	@Option(shortName = "DIAG", optional = true,
			doc="Name for a barcode match reporting file (not generated by default)."
					+"Either a name (in which case the file will be created in the output dir) or full path."
					+" This file will contain a line per read pair with the barcode best matching the read subsequence "
					+"or 'null' when no match is found accordign to matching parameters ; and the final selected sample."
					+" This file is useful for debugging or further processing in case both ends are barcoded.\n")
	String BARCODE_DIAG_FILE = null;
	File bcDiagFile = null; //final File object to use with BARCODE_DIAG_FILE, set in cmdline validation if necessary

	@Option(shortName = "O", optional = true,
			doc="Where to write output files. By default, these are written in running directory (see default value).\n")
	public File OUTPUT_DIR = new File(System.getProperty("user.dir"));

	@Option(optional = true,
			doc="Tells Jemultiplexer to overwrite files when already existing.\n" 
			)
	public boolean FORCE = false;
	
	
	@Option(shortName="GZ", optional = true,
			doc="Compress output s_l_t_barcode.txt files using gzip and append a .gz extension to the filenames.\n")
	public boolean GZIP_OUTPUTS = DEFAULT_GZIP_OUTPUTS;



	@Option(shortName="UN", optional = true,
			doc="Should un-assigned reads be saved in files or simply ignored. File names will be " +
					"automatically created or can be given using UF1 & UF2 options. Default names are "+DEFAULT_UNASSIGNED_FILE_NAME_1
					+" and "+DEFAULT_UNASSIGNED_FILE_NAME_2+" for F1 and F2 respectively.\n")
	public boolean KEEP_UNASSIGNED_READ = DEFAULT_KEEP_UNASSIGNED_READ;


	@Option(shortName="UF1", optional=true, 
			doc="Name of the file in which to write unassigned reads from FILE1."
					+"Either a name (in which case the file will be created in the output dir) or full path.\n"
			)
	public String UNASSIGNED_FILE_NAME_1 = DEFAULT_UNASSIGNED_FILE_NAME_1;
	public File unassignedFile1 = null; //final File object to use with UNASSIGNED_FILE1, set in cmdline validation if necessary

	@Option(shortName="UF2", optional=true,
			doc="Name of the file in which to write unassigned reads from FILE2."
					+"Either a name (in which case the file will be created in the output dir) or full path.\n"
			)
	public String UNASSIGNED_FILE_NAME_2 = DEFAULT_UNASSIGNED_FILE_NAME_2;
	public File unassignedFile2 = null;  //final File object to use with UNASSIGNED_FILE2, set in cmdline validation if necessary



	@Option(shortName="ASYNC", optional = true,
			doc="Use one thread per Fastq Writer.\n")
	public boolean WRITER_FACTORY_USE_ASYNC_IO = DEFAULT_WRITER_FACTORY_USE_ASYNC_IO;

	@Option(optional = true,
			doc="Only produces metric and diagnostic reports i.e. no output fastq file produced.\n")
	public boolean STATS_ONLY = false;



	private static final SolexaQualityConverter solexaQualityConverter = SolexaQualityConverter.getSingleton();



	/** Utility class to hang onto data about the best match for a given barcode */
	class BarcodeMatch {
		boolean matched;
		String barcode;
		int mismatches;
		int mismatchesToSecondBest;
		
		public String toString(){
			if(matched)
				return "matched :"+ barcode+" [MM="+mismatches+", MMD="+mismatchesToSecondBest+"]";
			return "no match";
		}
	}


	//sample name for unassigned read 
	private static final String UNASSIGNED = "UNASSIGNED";

	/*
	 * barcodes as byte arrays for both 
	 */
	protected byte[][] barcodeBytes1 = null;
	protected byte[][] barcodeBytes2 = null;

	public Jemultiplexer(){
		/*
		 * Loads properties
		 */
		try {
			ApplicationConfiguration.init();
		} catch (IOException e) {
			log.error("Failed to read properties from property file.");
			throw new RuntimeException(e);
		}
	}


	/**
	 * Validate that all BARCODEs are the same length and unique
	 *
	 * @return null if command line is valid.  If command line is invalid, returns an array of error message
	 *         to be written to the appropriate place.
	 */
	@Override
	protected String[] customCommandLineValidation() {
		final ArrayList<String> messages = new ArrayList<String>();


		/*
		 * Check mandatory input files
		 */
		if(!FASTQ_FILE1.exists())
			messages.add("FASTQ_FILE1 does not exist :"+FASTQ_FILE1.getAbsolutePath());

		if(FASTQ_FILE1.exists() && !FASTQ_FILE1.canRead())
			messages.add("Unsufficient rights to read FASTQ_FILE1 :"+FASTQ_FILE1.getAbsolutePath());

		if(FASTQ_FILE2 != null){
			RUNNING_PAIRED_END = true;

			if(!FASTQ_FILE2.exists())
				messages.add("FASTQ_FILE2 does not exist :"+FASTQ_FILE2.getAbsolutePath());

			if(FASTQ_FILE2.exists() &&!FASTQ_FILE2.canRead())
				messages.add("Unsufficient rights to read FASTQ_FILE2 :"+FASTQ_FILE2.getAbsolutePath());
			
			if(FASTQ_FILE2.getAbsolutePath().equals(FASTQ_FILE1.getAbsolutePath()))
				messages.add("FASTQ_FILE1 and FASTQ_FILE2 are the same file !");
		}
		
		if(INDEX_FILE1 != null){
			USE_SAMPLE_INDEX_FILES = true;
			
			
			if(!INDEX_FILE1.exists())
				messages.add("INDEX_FILE1 does not exist :"+INDEX_FILE1.getAbsolutePath());

			if(INDEX_FILE1.exists() &&!INDEX_FILE1.canRead())
				messages.add("Unsufficient rights to read INDEX_FILE1 :"+INDEX_FILE1.getAbsolutePath());
			
			if(INDEX_FILE1.getAbsolutePath().equals(FASTQ_FILE1.getAbsolutePath()))
				messages.add("FASTQ_FILE1 and INDEX_FILE1 are the same file !");
			
			if(RUNNING_PAIRED_END && INDEX_FILE1.getAbsolutePath().equals(FASTQ_FILE2.getAbsolutePath()))
				messages.add("FASTQ_FILE2 and INDEX_FILE1 are the same file !");
			
		}
		
		if(INDEX_FILE2 != null){
			USE_SAMPLE_INDEX_FILES = true;
			
			if(INDEX_FILE1 == null)
				messages.add("INDEX_FILE2 provided while INDEX_FILE1 was not given : can't give INDEX_FILE2 wihtout INDEX_FILE1");
			
			if(!INDEX_FILE2.exists())
				messages.add("INDEX_FILE2 does not exist :"+INDEX_FILE2.getAbsolutePath());

			if(INDEX_FILE2.exists() &&!INDEX_FILE2.canRead())
				messages.add("Unsufficient rights to read INDEX_FILE2 :"+INDEX_FILE2.getAbsolutePath());
			
			if(INDEX_FILE2.getAbsolutePath().equals(INDEX_FILE1.getAbsolutePath()))
				messages.add("INDEX_FILE1 and INDEX_FILE2 are the same file !");
			
			if(INDEX_FILE2.getAbsolutePath().equals(FASTQ_FILE1.getAbsolutePath()))
				messages.add("FASTQ_FILE1 and INDEX_FILE2 are the same file !");
			
			if(INDEX_FILE2.getAbsolutePath().equals(FASTQ_FILE2.getAbsolutePath()))
				messages.add("FASTQ_FILE2 and INDEX_FILE2 are the same file !");
			
			
		}

		if(USE_SAMPLE_INDEX_FILES){
			/*
			 * sample barcoding only occur in index file(s) when using index file(s)
			 */
			BARCODE_FOR_SAMPLE_MATCHING = BarcodePosition.NONE;
			
			/*
			 * if BPOS not given then turn it to NONE
			 */
			String _cmdline = Arrays.toString(this.getCommandLineParser().getArgv());
			log.debug(_cmdline);
			if(!_cmdline.contains("BPOS") && ! _cmdline.contains("BARCODE_READ_POS")){
				log.debug("[with sample index files] BPOS not given in cmd line : turning it to NONE");
				BARCODE_READ_POS = BarcodePosition.NONE;
			}
			
			/*
			 * if 2 index files were given and BRED not given, turn it to false
			 */
			if(INDEX_FILE2 !=null && !_cmdline.contains("BRED") && ! _cmdline.contains("REDUNDANT_BARCODES")){
				log.debug("[with sample index files] BRED not given in cmd line : turning it to false");
				REDUNDANT_BARCODES = false;
			}
			
			//BCLEN must be set to null unless BPOS != NONE as BCLEN is used for internal barcodes when using index files
			if(BARCODE_READ_POS == BarcodePosition.NONE){
				BCLEN = null;
			}
			
			
			if(!RUNNING_PAIRED_END){
				if(BARCODE_READ_POS==BarcodePosition.BOTH || BARCODE_READ_POS==BarcodePosition.READ_2){
					messages.add("[with sample index files] Can't have barcode(s) in position '"+BARCODE_READ_POS+"' in single end sequencing mode" );
				}
			}
			
			/*
			 * if barcodes are also in the reads, BLEN must be given
			 */
			if(BARCODE_READ_POS != BarcodePosition.NONE && net.sf.samtools.util.StringUtil.isBlank(BCLEN)){
				messages.add("[with sample index files] BCLEN must be given to describe the length of barcode(s) in read(s) : '"+BARCODE_READ_POS );
				return messages.toArray(new String[messages.size()]);
			}
				
			//change defaults of XT 
			if(!_cmdline.contains("XT") && !_cmdline.contains("XTRIMLEN")){
				//user did not specify this option, it must then be defaulted to 0 
				this.XTRIMLEN = "0";
			}
			
		}
		else{
			if(RUNNING_PAIRED_END){
				//if we have a unique barcode then BARCODE_FOR_SAMPLE_MATCHING value should match value of BARCODE_READ_POS
				if(BARCODE_READ_POS!=BarcodePosition.BOTH)
					BARCODE_FOR_SAMPLE_MATCHING = BARCODE_READ_POS;
				
			}
			else{
				BARCODE_READ_POS=BarcodePosition.READ_1;
				BARCODE_FOR_SAMPLE_MATCHING=BarcodePosition.READ_1;
			}
		}
		
		/*
		 * Intercept USE_EMBASE to initialize BF and other things
		 */
		if(USE_EMBASE){
			//use embase mode not supported with index files yet
			if(USE_SAMPLE_INDEX_FILES){
				messages.add("[with sample index files] USE_EMBASE mode not currently supported with index files");
				return messages.toArray(new String[messages.size()]);
			}
			messages.addAll( prepareForEmBASE() );
		}
		
		
		
		/*
		 * parse options that can have 'X:Y' syntax
		 * We first set these options blindly, later check is performed for those options that need it ie:
		 * - BCLEN* => make sure to have non 0 values at ends with barcodes and 0 values at ends w/o barcodes
		 * - XTRIMLEN => must be further inspected i.e. make sure we don t have a '1' for ends w/o 
		 * barcodes UNLESS explicitly set by user in the form X:Y 
		 */
		int [] arr = null;
		arr = setOption("BCLEN", BCLEN,  messages, true);
		if(arr.length == 2){ 
			//BCLEN was given in cmd line as it is null by default
			BCLEN_1 = arr[0];
			BCLEN_2 = arr[1];
		}else{
			//BCLEN was NOT given in cmd line ... let it init with barcode file...
		}
		
		
		
		arr = setOption("MAX_MISMATCHES", MAX_MISMATCHES,  messages, true);
		if(arr.length == 2){
			MAX_MISMATCHES_1 = arr[0];
			MAX_MISMATCHES_2 = arr[1];
		}
		
		arr = setOption("MIN_MISMATCH_DELTA", MIN_MISMATCH_DELTA, messages, true);
		if(arr.length == 2){
			MIN_MISMATCH_DELTA_1 = arr[0];
			MIN_MISMATCH_DELTA_2 = arr[1];
		}
		arr = setOption("MIN_BASE_QUALITY", MIN_BASE_QUALITY,  messages, true);
		if(arr.length == 2){
			MIN_BASE_QUALITY_1 = arr[0];
			MIN_BASE_QUALITY_2 = arr[1];
		}
		
		arr = setOption("XTRIMLEN", XTRIMLEN,  messages, false);
		if(arr.length == 2){
			XTRIMLEN_1 = arr[0];
			XTRIMLEN_2 = arr[1];
		}
		
		arr = setOption("ZTRIMLEN", ZTRIMLEN, messages, false);
		if(arr.length == 2){
			ZTRIMLEN_1 = arr[0];
			ZTRIMLEN_2 = arr[1];
		}
		
		//check XTRIMLEN (not needed when using INDEX_FILE as we already checked it before)
		if(!USE_SAMPLE_INDEX_FILES && RUNNING_PAIRED_END && BARCODE_READ_POS != BarcodePosition.BOTH){
			//set XTRIMLEN to 0 for the end w/o barcode UNLESS set by user in command line
			if(!XTRIMLEN.contains(":")){
				//user did not set 2 values => make it 0 
				if(BARCODE_READ_POS == BarcodePosition.READ_1)
					XTRIMLEN_2 = 0;
				else
					XTRIMLEN_1 = 0;
			}
		}
		
		
		
		
		/*
		 * Validate the barcode file format
		 */
		barcodeValidator = null; 
		if(BARCODE_FILE==null){
			if(USE_EMBASE)
				messages.add("Running EMBASE Mode: No Barcode file was created; this is most likely due to previous errors.");
			else
				messages.add("No Barcode file was given in command line!"); //this is normally NOT possible
		} 
		else if(!BARCODE_FILE.exists()){
			messages.add("BARCODE FILE does not exist :"+BARCODE_FILE.getAbsolutePath());
		}
		else if(!BARCODE_FILE.canRead()){
			messages.add("Unsufficient rights to read BARCODE FILE :"+BARCODE_FILE.getAbsolutePath());
		}
		else{
			barcodeValidator = new BarcodeValidator(BARCODE_FILE, this);
			//validate file format
			List<String> _messages = barcodeValidator.validate();
			/*
			 * when using index files, the lenght are barcodes found in the BC file has not been checked against what is found in the index files !
			 */
			if(USE_SAMPLE_INDEX_FILES){
				_messages.addAll(
						barcodeValidator.checkBarcodesLengthWithIndexFiles(INDEX_FILE1, INDEX_FILE2)
						);
			}
			
			if(_messages.size() != 0){
				messages.addAll(_messages);
			}
			else{
				
				/*
				 * first inspect a delicate situation: when we have 2 barcodes but only one is meant for barcode matching
				 * and the BC file does not have a bc1:bc2 syntax (which is expected in case the second barcode is 
				 * a random one) 
				 * in this situation, we must have BCLEN defined from cmd line for BOTH barcodes 
				 * 
				 * this situation is only possible when no index file is given 
				 * 
				 */
				if(	!USE_SAMPLE_INDEX_FILES 
						&& RUNNING_PAIRED_END 
						&& BARCODE_READ_POS==BarcodePosition.BOTH 
						&& !REDUNDANT_BARCODES && BARCODE_FOR_SAMPLE_MATCHING!= BarcodePosition.BOTH 
						&& !barcodeValidator.hasBarcodesForBothEnds ){
					 
					if(BCLEN_1 == 0 && BCLEN_2 == 0){
						 //we dont fail ONLY if CLIP barcode is false and ADD is false as well
						 if(CLIP_BARCODE || ADD_BARCODE_TO_HEADER){
							 messages.add("BCLEN for BOTH barcode must be specified in the command line (e.g. BCLEN=6:8) when" +
							 		" BARCODE_READ_POS==BOTH REDUNDANT_BARCODES=false BARCODE_FOR_SAMPLE_MATCHING="+BARCODE_FOR_SAMPLE_MATCHING.toString()+
							 		" and the barcode file does not distinguish between the barcodes of both ends (using bc1:bc2 syntax). " +
							 		"The situation is too ambigous.");
						 }
					 }
					
				}
				
				
				/*
				 * set barcode length if needed ; if still equals to 0 => BCLEN option was not
				 * given in cmd line ; get value from barcodeValidator if applicable 
				 */
				if(BCLEN_1 == 0){  
					if(!USE_SAMPLE_INDEX_FILES && (!RUNNING_PAIRED_END || BARCODE_READ_POS!=BarcodePosition.READ_2))
						BCLEN_1 = barcodeValidator.bclen1;
					else if (USE_SAMPLE_INDEX_FILES && (BARCODE_READ_POS == BarcodePosition.READ_1 || BARCODE_READ_POS == BarcodePosition.BOTH)){
						messages.add("[using index file] barcode lenght for READ 1 cannot be 0 with BPOS="+BARCODE_READ_POS);
					}
				}
				else if (BCLEN_1 != 0 && RUNNING_PAIRED_END && BARCODE_READ_POS==BarcodePosition.READ_2){
					//it should be 0 if no barcode is described at this end
					BCLEN_1 = 0;
				}

				
				/*
				 * set barcode length if needed ; if still equals to 0 => BCLEN option was not given 
				 * in cmd line ; get value from barcodeValidator if applicable
				 */
				if(BCLEN_2 == 0 && RUNNING_PAIRED_END && (BARCODE_READ_POS==BarcodePosition.READ_2 || BARCODE_READ_POS == BarcodePosition.BOTH)){ 
					if(USE_SAMPLE_INDEX_FILES && (BARCODE_READ_POS == BarcodePosition.READ_2 || BARCODE_READ_POS == BarcodePosition.BOTH)){
						messages.add("[using index file] barcode lenght for READ 2 cannot be 0 with BPOS="+BARCODE_READ_POS);
					}else if(!USE_SAMPLE_INDEX_FILES){
						BCLEN_2 = barcodeValidator.bclen2;
					}
				}
				else if (BCLEN_2 != 0 && RUNNING_PAIRED_END && BARCODE_READ_POS==BarcodePosition.READ_1){
					//it should be 0 if no barcode is described at this end
					BCLEN_2 = 0;
				}

				//more over, it does not make sense to have different BCLEN if barcodes are "redundant" (only when not using index files)
				if(!USE_SAMPLE_INDEX_FILES && RUNNING_PAIRED_END 
						&& BARCODE_READ_POS==BarcodePosition.BOTH && BARCODE_FOR_SAMPLE_MATCHING==BarcodePosition.BOTH 
						&& REDUNDANT_BARCODES && BCLEN_1 != BCLEN_2 ){
					messages.add("It does not make sense to have different barcode length (BCLEN1="+BCLEN_1+", BCLEN2="+BCLEN_2+") "
							+" when REDUNDANT_BARCODES=true ");
				}
			}
		}
		

		/*
		 * check output dir
		 */
		if(!USE_EMBASE && !OUTPUT_DIR.exists()){
			log.info("Attempting to create output directory : "+OUTPUT_DIR.getAbsolutePath());
			FileUtil.checkWritableDir(OUTPUT_DIR, true);
		}

		/*
		 * Check quality format
		 */
		if (QUALITY_FORMAT == null) {
			final QualityEncodingDetector detector = new QualityEncodingDetector();
			final FastqReader reader = new FastqReader(FASTQ_FILE1);
			if (FASTQ_FILE2 == null) {
				detector.add(QualityEncodingDetector.DEFAULT_MAX_RECORDS_TO_ITERATE, reader);
			} else {
				final FastqReader reader2 = new FastqReader(FASTQ_FILE2);       
				detector.add(QualityEncodingDetector.DEFAULT_MAX_RECORDS_TO_ITERATE, reader, reader2);
				reader2.close();
			}
			reader.close();

			QUALITY_FORMAT = detector.generateBestGuess(QualityEncodingDetector.FileContext.FASTQ);
			if (detector.isDeterminationAmbiguous())
				log.warn("Making ambiguous determination about fastq's quality encoding; more than one format possible based on observed qualities.");

			log.info(String.format("Auto-detected quality format as: %s.", QUALITY_FORMAT));
		}

		if(READ_NAME_REPLACE_CHAR!=null){
			if(!StringUtil.isValid(READ_NAME_REPLACE_CHAR)){
				messages.add("The option READ_NAME_REPLACE_CHAR is empty, for special characters like ';', you might need to escape them with a backslash.");
			}
			log.debug("RCHAR is "+READ_NAME_REPLACE_CHAR);
		}

		if (messages.size() == 0) {
			/*
			 * All is ok, finish understanding the command line options 
			 */

			if(METRICS_FILE_NAME == null) METRICS_FILE_NAME=DEFAULT_METRICS_FILE_NAME; //this is user mistake, reset the default name
			metricsFile = setFile(METRICS_FILE_NAME);

			if(BARCODE_DIAG_FILE != null)
				bcDiagFile = setFile(this.BARCODE_DIAG_FILE);

			if(KEEP_UNASSIGNED_READ){
				if(UNASSIGNED_FILE_NAME_1 == null) UNASSIGNED_FILE_NAME_1=DEFAULT_UNASSIGNED_FILE_NAME_1; //this is user mistake, reset the default name
				if(UNASSIGNED_FILE_NAME_2 == null && RUNNING_PAIRED_END) UNASSIGNED_FILE_NAME_2=DEFAULT_UNASSIGNED_FILE_NAME_2; //this is user mistake, reset the default name

				if(GZIP_OUTPUTS && ! UNASSIGNED_FILE_NAME_1.endsWith(".gz"))
					UNASSIGNED_FILE_NAME_1 += ".gz";

				unassignedFile1 = setFile(UNASSIGNED_FILE_NAME_1);
				if(RUNNING_PAIRED_END){
					if(GZIP_OUTPUTS && ! UNASSIGNED_FILE_NAME_2.endsWith(".gz"))
						UNASSIGNED_FILE_NAME_2 += ".gz";
					unassignedFile2 = setFile(UNASSIGNED_FILE_NAME_2);
				}
			}

			
			return null;
		}
		return messages.toArray(new String[messages.size()]);
	}



	
	

	/** 
	 * Called at cmd line validation when the USE_EMBASE switch is used 
	 * @param messages
	 */
	protected List<String> prepareForEmBASE() {
		//override whatever user could have set for some options
		//
		
		//error messages to be accumulated
		List<String> messages = new ArrayList<String>();
		
		
		OUTPUT_DIR = null;
		//we keep unassigned reads
		KEEP_UNASSIGNED_READ = true;
		//and call files accordign to input files, simply adding '_unassigned-reads' before extension
		UNASSIGNED_FILE_NAME_1= FileUtil.removeExtension( FASTQ_FILE1.getName() ) + "_unassigned-reads"+".txt.gz" ;
		UNASSIGNED_FILE_NAME_2= RUNNING_PAIRED_END ? FileUtil.removeExtension( FASTQ_FILE2.getName() ) + "_unassigned-reads"+".txt.gz" : null;
		GZIP_OUTPUTS=true;
		STATS_ONLY=false;
		CREATE_MD5_FILE = true;
		
		/*
		 * Get information from emBASE 
		 */		
		
		//Determine the caller
		String username = System.getProperty("user.name");
		//check user is not trying to trick me with a -Duser.name=another user!
		String whoami = whoami();
		if(!username.equalsIgnoreCase(whoami)){
			messages.add("User "+username+" does not match 'whoami' ("+whoami+") ! Are you trying to hack the system ?");
			//we cannot continue
			return messages;
		}
		
		EmBaseDatabaseFacade facade = null;
		try {
			facade = EmBaseDatabaseFacade.createEmBaseDatabaseFacade(
					username,
					ApplicationConfiguration.DB_URL,
					ApplicationConfiguration.DB_USER,
					ApplicationConfiguration.DB_PWD
					);
		} catch (EmBASEConnectionException e) {
			log.error(ExceptionUtil.getStackTrace(e));
			messages.add("Error when trying to connect emBASE DB : "+e.getMessage());
			//we cannot continue
			return messages;
		}
		
		//what s the lane number ? first get teh NGS assay
		Item ngsAssay = facade.getNGSAssay(FASTQ_FILE1);
		List<CSVLine> rbaDetails = null;
		if(ngsAssay == null){
			messages.add("No NGS Assay found in emBASE for lane file : "+FASTQ_FILE1.getAbsolutePath()+"\nAre you sure this file has been already imported in emBASE ? If so, please contact emBASE admins.");
		}else{
			//then fetch details
			ArrayList<Item> _list = new ArrayList<Item>();
			_list.add(ngsAssay);
			/*
			 * rbaDetails contains values for all org.embl.gbcs.embase.api.Queries.HEADER_*
			 */
			rbaDetails = facade.fetchRBADetailsForNGSAssays(_list).values().iterator().next();
			if(rbaDetails == null || rbaDetails.size()==0){
				messages.add("No RawBioAssay found in emBASE for NGS Assay "+ngsAssay.getName()+" (id="+ngsAssay.getId()+") i.e. lane file : "+FASTQ_FILE1.getAbsolutePath()+"\nAre you sure this file has been successfully imported in emBASE ? If so, please contact emBASE admins.");
			}
		}
		
		if(messages.size()!=0){
			//we stop here
			return messages;
		}
		
		//get lane number : any RBA can be used for this
		Integer laneNumber = null;
		try {
			laneNumber = rbaDetails.get(0).getValueAsInt(Queries.HEADER_Lane);
		} catch (InvalidHeaderException e1) {
			//not possible if the code has been run at least once!
			throw new RuntimeException(e1);
		}
		
		// Check that expected directories exist
		File runDir = FASTQ_FILE1.getParentFile();
		//get lane dir
		File laneDir = new File (runDir, EmBaseDatabaseFacade.getLaneDirName(laneNumber));
		//then check if expected dir exists for stats and unassigned reads
		File runStatDir = new File (laneDir, EmBaseDatabaseFacade.STAT_DIR_NAME);
		
		if(!runStatDir.exists() || !runStatDir.canWrite()){
			if(!TEST_MODE_STOP_AFTER_PARSING)
				messages.add("The expected '"+EmBaseDatabaseFacade.STAT_DIR_NAME+"' dir in the lane dir '"+laneDir.getAbsolutePath()+"' is either not existing or not writable");
			else{
				METRICS_FILE_NAME = new File(EmBaseDatabaseFacade.TMP_DIR, FileUtil.removeExtension( FASTQ_FILE1.getName() ) +"_jemultiplexing_metrics.txt").getAbsolutePath();
				log.info("TEST MODE: METRICS_FILE_NAME set to "+METRICS_FILE_NAME);
			}
		}else{
			//set the full path so the remaining of the code leaves it like this
			METRICS_FILE_NAME = new File(runStatDir, FileUtil.removeExtension( FASTQ_FILE1.getName() ) +"_jemultiplexing_metrics.txt").getAbsolutePath();
			log.info("METRICS_FILE_NAME set to "+METRICS_FILE_NAME);
		}
		
		File unassignedReadsDir = new File (laneDir, EmBaseDatabaseFacade.UNASSIGNED_DIR_NAME);
		if(!unassignedReadsDir.exists() || !unassignedReadsDir.canWrite()){
			if(!TEST_MODE_STOP_AFTER_PARSING)
				messages.add("The expected '"+EmBaseDatabaseFacade.UNASSIGNED_DIR_NAME+"' dir in the run dir '"+laneDir.getAbsolutePath()+"' is either not existing or not writable");
			else{
				OUTPUT_DIR = EmBaseDatabaseFacade.TMP_DIR;
				log.info("TEST MODE: OUTPUT_DIR set to "+OUTPUT_DIR.getAbsolutePath());
			}
		}else{
			 //set output dir to where the input files are located => this will be the place for unassigned reads 
			OUTPUT_DIR = unassignedReadsDir;
			log.info("OUTPUT_DIR set to "+OUTPUT_DIR.getAbsolutePath());
		}

		
		/*
		 * Check caller can actually demultiplex this dataset
		 * The user should have read access on the NGS assay
		 */
		try {
			if(!facade.canDemultiplex(FASTQ_FILE1, username)){
				messages.add("According to emBASE DB, the user "+username+" is not allowed to read NGS Assay corresponding to file "+FASTQ_FILE1.getAbsolutePath());
			}else{
				log.debug("User allowed to demultiplex");
			}
		}catch (Exception e) {
			log.error(ExceptionUtil.getStackTrace(e));
			messages.add("Error when trying to check read rights from emBASE DB : "+e.getMessage());
		}
		
		
		/*
		 * Get all libs 
		 */
		List<NGSLibrary> libraries = null;
		try {
			libraries = facade.fetchLibraries(FASTQ_FILE1);
			facade.findSampleStorageDir(laneDir, libraries);
			
			//and check dirs have been set for all libs
			for (NGSLibrary lib : libraries) {
				if(lib.getSampleDir() == null || !lib.getSampleDir().exists()){
					String mess = "No library dir for sample/library "+lib.getName()+" is available.\nThe emBASE storage architecture is not ready. " +
							"Are you sure this lane has been successfully uploaded to emBASE ?\nIf so, please contact emBASE admins.";
					log.debug(mess);
					if(TEST_MODE_STOP_AFTER_PARSING){
						log.debug("TEST MODE : Setting fake dir for "+lib.getName());
						lib.setSampleDir ( EmBaseDatabaseFacade.getFakeSampleDir(lib.getName(), lib.getId()) );//does not exist, just fake for test
					}
					else{
						messages.add(mess);
					}
				}
				else {
					//update lib.getSampleDir with fastq sub-dir
					File fastqDir = new File(lib.getSampleDir(), "fastq");
					if(!fastqDir.exists() || !fastqDir.canWrite()){
						String mess = "The 'fastq' dir under the library dir  "+lib.getSampleDir().getAbsolutePath()
								+" is either not existing OR not writable.\n";
						if(!fastqDir.exists()){
							mess+="The storage architecture is not ready. Are you sure this lane has been successfully uploaded to emBASE ?\nIf so, please contact emBASE admins.";
						}
						else{
							mess+="The fastq storage file has been already locked to prevent further modifications.\n" +
									"Please contact emBASE admins if you think this should not be the case.";
						}
						log.debug(mess);
						messages.add(mess);
					}else{
						lib.setSampleDir( fastqDir );
					}
				}
			}
		} catch (Exception e) {
			log.error(ExceptionUtil.getStackTrace(e));
			messages.add("Error when trying to fetch barcodes from emBASE DB : "+e.getMessage());
		}
		
		if(messages.size()!=0){
			//we stop here
			return messages;
		}
		
		/*
		 * All seems perfect, create barcode file 
		 * Check demultiplexed files are not already found
		 */
		PrintWriter pw = null;
		try{
			BARCODE_FILE = getBarcodeFileTmpLocation(FASTQ_FILE1, username, runStatDir);
			if(BARCODE_FILE.exists()){
				log.warn("The barcode file already exists and will be overwritten : "+BARCODE_FILE.getAbsolutePath());
			}
			pw = new PrintWriter(BARCODE_FILE);
			for (NGSLibrary lib : libraries) {
				//File fastq1 = getDemultiplexedFile(lib, FileUtil.removeExtension(FASTQ_FILE1.getName()) , GZIP_OUTPUTS);
				//File fastq2 = RUNNING_PAIRED_END ? getDemultiplexedFile(lib, FileUtil.removeExtension(FASTQ_FILE2.getName()), GZIP_OUTPUTS) : null;
				
				File fastq1 = getDemultiplexedFile(lib, 1 , GZIP_OUTPUTS);
				File fastq2 = getDemultiplexedFile(lib, 2 , GZIP_OUTPUTS);
				
				
				/*
				 * check if file already exists , the test on the length is meant to avoid considering a 'fake' file as a real file
				 * we indeed create such files to prepare the architecture
				 */
				if(fastq1.exists() && fastq1.length() > 10*1000L && !FORCE){ 
					if(fastq1.canWrite()){
						messages.add("Demultiplexed file "+fastq1.getAbsolutePath()+" already exists. You need to run with FORCE=true to overwrite existing files.");
					}else{
						messages.add("Demultiplexed file "+fastq1.getAbsolutePath()+" already exists BUT you do NOT have write permission on it.\n" +
								"You first need to solve the write permission issue, then use FORCE=true to overwrite existing files.");
					}
				}
				else if(fastq1.exists() && FORCE && fastq1.canWrite()){
					log.warn("Existing demultiplexed file will be overritten (option FORCE=true) : "+fastq1.getAbsolutePath());
				}
				
				if(RUNNING_PAIRED_END && fastq2.exists() & fastq2.length() > 10*1000L && !FORCE){
					if(fastq2.canWrite()){
						messages.add("Demultiplexed file "+fastq2.getAbsolutePath()+" already exists. You need to run with FORCE=true to overwrite existing files.");
					}else{
						messages.add("Demultiplexed file "+fastq2.getAbsolutePath()+" already exists BUT you do NOT have write permission on it.\n" +
								"You first need to solve the write permission issue, then use FORCE=true to overwrite existing files.");
					}
				}else if(RUNNING_PAIRED_END && fastq2.exists() && FORCE && fastq2.canWrite()){
					log.warn("Existing demultiplexed file will be overritten (option FORCE=true) : "+fastq2.getAbsolutePath());
				}
				
				pw.print(lib.getName()+"\t"+lib.getBarcode()+"\t"+fastq1.getAbsolutePath());
				if(RUNNING_PAIRED_END)
					pw.print("\t"+fastq2.getAbsolutePath());
				pw.println();
			}
		} catch (FileNotFoundException e) {
			throw new RuntimeException(e);
		}finally{
			if(pw!=null)
				pw.close();
		}
		
		//then ok just let the code go on
		
		return messages;
	}


	
	/**
	 * builds a file to use for barcode file
	 * 
	 * @param fastqFile the fastq file (mate 1 in case of PE) 
	 * @param username the user how is about to demultiplex
	 * @param d the dir to use or null to use system defaults
	 * @return
	 */
	protected File getBarcodeFileTmpLocation(File fastqFile, String username, File d) {
		String tmpdir = d == null ? System.getProperty("java.io.tmpdir") : d.getAbsolutePath();
		if(!StringUtil.isValid(tmpdir) || !new File(tmpdir).exists())
			tmpdir = ".";
		
		File f= new File(
				tmpdir,
				FileUtil.removeExtension( fastqFile.getName() ) +"_"+username+"_barcodes.txt"
				);
		log.debug("barcode file to use is "+f.getAbsolutePath());
		return f;
	}
	

	/**
	 * Deals with integer options that can take X:Y values ; also check that given values are >= 0 
	 * @param optionName name of the option (for error messages)
	 * @param optionValue as given by user ; if blank returned array is empty
	 * @param messages the error message list to use in case of error
	 * @param failIfPEwithSingleBarcode
	 * @return an int array with the two parsed values for read_1 and read_2. 
	 * Array is empty is nothing should be made ie in such case one should check if messages size increased
	 */
	private int[] setOption(String optionName, String optionValue, ArrayList<String> messages, boolean failIfPEwithSingleBarcode) {
		
		int[] arr = new int[2];
		
		if(net.sf.samtools.util.StringUtil.isBlank(optionValue))
			return arr;
		
		String [] values = new String[2]; 
		if(optionValue.contains(":")){
			if(USE_SAMPLE_INDEX_FILES && INDEX_FILE2 == null && failIfPEwithSingleBarcode && !optionName.equalsIgnoreCase("BCLEN")){
				messages.add("Invalid value used for "+optionName+" : "+optionValue +". The X:Z synthax can only be used with 2 sample index files");
				return arr ;
			}
			else if(!RUNNING_PAIRED_END || (failIfPEwithSingleBarcode && BARCODE_READ_POS!=BarcodePosition.BOTH)){
				messages.add("Invalid value used for "+optionName+" : "+optionValue +". The X:Z synthax can only be used for paired-end data with BPOS=BOTH");
				return arr ;
			}
			//split values
			values = optionValue.split(":");
			if(values.length!=2){
				messages.add("Invalid value used for "+optionName+" : "+optionValue +". The X:Z synthax can contain only a single ':' ");
				return arr;
			}
		}
		else{
			values[0] = optionValue;
			values[1] = optionValue;
		}
		
		//now convert to int
		int optionRead1 = -1; 
		int optionRead2 = -1;
		try{
			optionRead1 = Integer.parseInt(values[0]);
		}catch (NumberFormatException e) {
			messages.add("Invalid value used for "+optionName+" : "+optionValue +" => "+values[0]+" is not an integer");
		}
		
		try{
			optionRead2 = Integer.parseInt(values[1]);
		}catch (NumberFormatException e) {
			messages.add("Invalid value used for "+optionName+" : "+optionValue +" => "+values[1]+" is not an integer");
		}
		
		log.debug(optionName+" => [1]="+optionRead1+" ; [2]="+optionRead2);
		
		if(optionRead1<0 || optionRead2<0)
			messages.add("Invalid value used for "+optionName+" : "+optionValue +" => value(s) must be >= 0.");
		
		if(RUNNING_PAIRED_END && BARCODE_READ_POS==BarcodePosition.BOTH && REDUNDANT_BARCODES && optionRead1 != optionRead2 ){
			log.warn("POTENTIAL OPTION MISTAKE : Found 2 different values for option "+optionName+" (given cmd line value was '"+optionValue
					+"') while REDUNDANT_BARCODES=true: "+optionRead1+" for read_1 and "+optionRead2+" for read_2. Code continues for now.");
		}
		
		return new int []{optionRead1, optionRead2};
	}


	private File setFile(String fnameOrPath) {
		if(fnameOrPath.contains("/")){
			//it is a path, use file path as provided
			return new File(fnameOrPath);
		}
		else{
			//it is just a name
			return new File(OUTPUT_DIR, fnameOrPath);
		}

	}

	
	
	
	/**
	 * 
	 * Find out what file should be used for a given library. This method does NOT fetches existing file
	 * but builds a file path to use by convention i.e. when you want to automatically demultiplex fastq file
	 *  
	 * 
	 * @param lib the lib for which we need to know the file to use to write demultiplexed reads.
	 * Note that lib should have its final sampleDir correctly set 
	 * @param readPairNumber 1 or 2 (for single end, it is always 1) 
	 * @param gzip whether output should be gzipped
	 * @return the File to use
	 */
	public File getDemultiplexedFile(NGSLibrary lib, int readPairNumber, boolean gzip) {
		
		//for now simply build up a name with lib name
		File f =  new File(lib.getSampleDir(), lib.getName()+"_"+lib.getSampleDir().getParentFile().getName()+"_"+readPairNumber+".txt"+(gzip?".gz":"")); 
		log.debug("Demultiplexed file for "+lib.getName()+" (read pair "+readPairNumber+") should be "+f.getAbsolutePath());
		return f;
	}

	@SuppressWarnings("unused")
	@Override
	protected int doWork() {

		
		if(TEST_MODE_STOP_AFTER_PARSING)
			return 0;
		
		//should we write a barcode match result file ?
		boolean writingBarcodeDiagFile = (this.bcDiagFile != null);

		
		/*
		 * Initialize all barcode maps 
		 */
		barcodeValidator.parse();
	
		Map<String, String> bc2sample = barcodeValidator.barcode2sample;
		Map<String, Set<String>> sample2bcSet = barcodeValidator.sample2barcodes;
		//fill in barcodeBytes array
		List<String> barcodeList1 = new ArrayList<String>(barcodeValidator.barcodeSetRead1);
		barcodeBytes1 = new byte[barcodeList1.size()][];
		for (int i = 0; i < barcodeList1.size(); i++) {
			barcodeBytes1[i] = net.sf.samtools.util.StringUtil.stringToBytes(barcodeList1.get(i));
		}
		List<String> barcodeList2 = new ArrayList<String>(barcodeValidator.barcodeSetRead2);
		barcodeBytes2 = new byte[barcodeList2.size()][];
		for (int i = 0; i < barcodeList2.size(); i++) {
			barcodeBytes2[i] = net.sf.samtools.util.StringUtil.stringToBytes(barcodeList2.get(i));
		}

		/*
		 * prepare FASTQ file writers
		 */
		JemultiplexerFastqWriterFactory fastqFactory = new JemultiplexerFastqWriterFactory();
		fastqFactory.setUseAsyncIo(WRITER_FACTORY_USE_ASYNC_IO);

		//init a writer and counter for each sample
		HashMap<String, FastqWriter> writers_fwd = new HashMap<String, FastqWriter>();
		HashMap<String, FastqWriter> writers_rev = new HashMap<String, FastqWriter>();

		//map storing how many read pair are assigned to each sample
		HashMap<String, Integer> sampleCountMap = new HashMap<String, Integer>(); 

		//do we use long barcode format?
		Map<String, String> sample2file1 = null;
		Map<String, String> sample2file2 = null;
		if(barcodeValidator.isLongFormat && !STATS_ONLY){
			//yes load file name to use
			sample2file1 = barcodeValidator.sample2file1;
			if(RUNNING_PAIRED_END)
				sample2file2 = barcodeValidator.sample2file2;
		}
		for (Entry<String, Set<String>> e : sample2bcSet.entrySet()) {
			String sample = e.getKey();
			sampleCountMap.put(sample, 0);
			List<String> barcodes = new ArrayList<String>( e.getValue() );
			//default names
			String fname_fwd = sample + "_"+StringUtil.mergeList(barcodes, "-")+"_1.txt";
			String fname_rev = sample + "_"+StringUtil.mergeList(barcodes, "-")+"_2.txt";
			//overwrite with user provided file names if available
			fname_fwd = sample2file1 == null || !sample2file1.containsKey(sample) ? fname_fwd : sample2file1.get(sample);
			fname_rev = sample2file2 == null || !sample2file2.containsKey(sample) ? fname_rev : sample2file2.get(sample);
			if(GZIP_OUTPUTS){
				//make sure files ends with .gz
				if(!fname_fwd.endsWith(".gz"))
					fname_fwd += ".gz";
				if(fname_rev!=null && !fname_rev.endsWith(".gz"))
					fname_rev += ".gz";
			}
			if(!STATS_ONLY){
				log.info("Registering writers for "+sample+" :");
				log.info("\tFile 1 Name : "+fname_fwd);
				if(isAbsolutePath(fname_fwd)){
					writers_fwd.put(sample, fastqFactory.newWriter(new File(fname_fwd), GZIP_OUTPUTS, CREATE_MD5_FILE));
				}else{
					log.info("OUTPUT_DIR:"+OUTPUT_DIR.getAbsolutePath());
					writers_fwd.put(sample, fastqFactory.newWriter(new File(OUTPUT_DIR, fname_fwd), GZIP_OUTPUTS, CREATE_MD5_FILE));
				}
				
				if(RUNNING_PAIRED_END){
					log.info("\tFile 2 Name : "+fname_rev);
					if(isAbsolutePath(fname_rev)){
						writers_rev.put(sample, fastqFactory.newWriter(new File(fname_rev), GZIP_OUTPUTS, CREATE_MD5_FILE));
					}else{
						writers_rev.put(sample, fastqFactory.newWriter(new File(OUTPUT_DIR, fname_rev), GZIP_OUTPUTS, CREATE_MD5_FILE));
					}
					
				}
			}
		}
		//add writers for the unassigned reads if required
		if(KEEP_UNASSIGNED_READ && !STATS_ONLY){
			log.info("Registering writers unassigned reads :");
			log.info("\tFile 1 Name : "+unassignedFile1);
			writers_fwd.put(UNASSIGNED, fastqFactory.newWriter(unassignedFile1, GZIP_OUTPUTS, CREATE_MD5_FILE));
			if(RUNNING_PAIRED_END){
				writers_rev.put(UNASSIGNED, fastqFactory.newWriter(unassignedFile2, GZIP_OUTPUTS, CREATE_MD5_FILE));
				log.info("\tFile 2 Name : "+unassignedFile2);
			}
		}

		/*
		 * Should we init the diag file ?
		 */
		PrintWriter diagPW = null;
		if( writingBarcodeDiagFile ){
			try{
				diagPW = new PrintWriter(bcDiagFile);
			} catch (Exception e) {
				log.error(ExceptionUtil.getStackTrace(e));
				throw new RuntimeException(e);
			}
			/*
			 * init headers with SE  :
			 * - Read_Header (is read header as found in FASTQ) ; 
			 * - "seq" : read sub-sequence to match against barcodes ; 
			 * - "MM_Best", "MM_Second": Mismatch number with best and second best match
			 * - "bc" : chosen barcode sequence ; this is ofcourse function of global options
			 * - "sample" : sample ultimately selected ie the read should be found in the corresponding file
			 * For PE case only :
			 * - same as above but duplicated with ending _1 and _2 for read 1 and 2, respectively ; 
			 * - "smpl_1" and "smpl_2" : sample match for bc1 and bc2
			 */
			String[] headers = new String []{"Read_Header", "seq", "MM_Best", "MM_Second", "bc", "sample"};
			if(RUNNING_PAIRED_END)
				headers = new String []{"Read_1_Header", "Read_2_Header", "seq1", "seq2" , "MM_Best_1", "MM_Second_1", "bc1", "smpl_1", "MM_Best_2", "MM_Second_2", "bc2", "smpl_2", "sample"};

			diagPW.println(StringUtil.mergeArray(headers,"\t"));
		}

		FastqReader fqr1 = new FastqReader(FASTQ_FILE1);
		Iterator<FastqRecord> it1 = fqr1.iterator();
		FastqReader fqr2 = null;
		Iterator<FastqRecord> it2 = null;

		FastqReader idxr1 = null;
		Iterator<FastqRecord> idxit1 = null;

		FastqReader idxr2 = null;
		Iterator<FastqRecord> idxit2 = null;

		
		if(RUNNING_PAIRED_END){
			fqr2 = new FastqReader(FASTQ_FILE2);
			it2 = fqr2.iterator();
		}

		if(USE_SAMPLE_INDEX_FILES){
			idxr1 = new FastqReader(INDEX_FILE1);
			idxit1 = idxr1.iterator();
			
			if(INDEX_FILE2!=null){
				idxr2 = new FastqReader(INDEX_FILE2);
				idxit2 = idxr2.iterator();
			}
		}
		
		/*
		 * Read files and write
		 */
		int cnt = 0; //read pair counter
		int unassigned = 0; //count how many pairs end up unassigned 
		int assigned = 0; //count how many pairs end up with a sample assignment
		while(it1.hasNext()){
			FastqRecord r1 = it1.next();
			FastqRecord r2 = null; 
			if(RUNNING_PAIRED_END){
				r2 = it2.next();
			}
			cnt++;
			if(cnt % (1000*1000) == 0){
				log.info(cnt+" reads processed");

			}

			/*
			 * Operate barcode EXTRACTION 
			 */
			String s1 = null;
			String q1 = null;
			BarcodeMatch bc1 = null;
			String s2 = null;
			String q2 = null;
			BarcodeMatch bc2 = null;
			
			if(USE_SAMPLE_INDEX_FILES){
				FastqRecord idx_r1 = idxit1.next();
				s1 = idx_r1.getReadString();
				q1 = idx_r1.getBaseQualityString();
				log.debug("BC1="+s1);
				if(INDEX_FILE2!=null){
					FastqRecord idx_r2 = idxit2.next();
					s2 = idx_r2.getReadString();
					q2 = idx_r2.getBaseQualityString();
					log.debug("BC2="+s2);
				}
				
			}
			else{
				//do we have a barcode at read 1 : yes if SE or PE with BC position != read2 
				if(!RUNNING_PAIRED_END || BARCODE_READ_POS != BarcodePosition.READ_2){
					//read 1 has a barcode, extract it
					s1 = r1.getReadString().substring(0, BCLEN_1).toUpperCase();
					q1 = r1.getBaseQualityString().substring(0, BCLEN_1);
				}
				//do we have a barcode at read 2 ? => yes is PE with BC position != read1 
				if(RUNNING_PAIRED_END && BARCODE_READ_POS != BarcodePosition.READ_1){
					s2 = r2.getReadString().substring(0, BCLEN_2);
					q2 = r2.getBaseQualityString().substring(0, BCLEN_2);
				}
			}
			
			
			
			/*
			 * Operate barcode MATCHING
			 */
			//do we need to match barcode 1? 
			//yes if (1) single end or (2) there is a barcode in READ1 position AND it is used for sample matching
			// or (3) we use index file
			if(USE_SAMPLE_INDEX_FILES || !RUNNING_PAIRED_END || (BARCODE_READ_POS != BarcodePosition.READ_2 && BARCODE_FOR_SAMPLE_MATCHING != BarcodePosition.READ_2 )){
				bc1 = findBarcode(BarcodePosition.READ_1, s1, q1, bc2sample, sample2bcSet);
				log.debug("bc1 match : "+ (bc1 == null?"NULL":bc1.toString()));
			}
			
			//do we need to match barcode 2? 
			// yes if (1) running paired-end AND there is a barcode in READ2 position AND it is used for sample matching
			// or (2) we use 2 index files
			if(	(USE_SAMPLE_INDEX_FILES && INDEX_FILE2!=null) 
					|| (RUNNING_PAIRED_END && BARCODE_READ_POS != BarcodePosition.READ_1 && BARCODE_FOR_SAMPLE_MATCHING != BarcodePosition.READ_1)){
				bc2 = findBarcode(BarcodePosition.READ_2, s2, q2, bc2sample, sample2bcSet);
				log.debug("bc2 match : "+ (bc2 == null?"NULL":bc2.toString()));
			}
			
			
			
			/*
			 * Special case when not using INDEX FILES
			 *  we have 2 barcode but only one is used for sample lookup, the second one being most likely a random tag 
			 * for PCR duplicate assessment => we still need to fill in bc1 or bc2
			 * 
			 * Indeed, if index files are used then bc are set correctly and we can t be in a weird situation (also
			 * remember that using index files is exclusive ie no barcode for sample matching can be in the reads
			 *  anymore, only molecular indices)
			 * 
			 */
			if(!USE_SAMPLE_INDEX_FILES){
				if(RUNNING_PAIRED_END && BARCODE_READ_POS == BarcodePosition.BOTH && BARCODE_FOR_SAMPLE_MATCHING != BarcodePosition.BOTH){
					if(BARCODE_FOR_SAMPLE_MATCHING == BarcodePosition.READ_1){
						//set bc2
						bc2 = new BarcodeMatch();
						bc2.matched = true;
						bc2.barcode = s2;
						bc2.mismatches = 0;
						bc2.mismatchesToSecondBest = s2.length();
					}
					else{
						bc1 = new BarcodeMatch();
						bc1.matched = true;
						bc1.barcode = s1;
						bc1.mismatches = 0;
						bc1.mismatchesToSecondBest = s1.length();
					}
				}
			}

			
			
			
			/*
			 * Operate the barcode to sample matching
			 * At this point writers are set to null. Writers will be set only if barcodes
			 * can be mapped to a sample. 
			 */
			FastqWriter fwdW = null;
			FastqWriter revW = null;
			String spl1 = null; //the sample read_1 resolves to
			String spl2 = null; //the sample read_2 resolves to
			String chosen_sample = null; //final sample to attribute the read to
	
			
			/*
			 * Sample Lookup : set the fwdW, revW writers and return an array with samples variables (in this order) : spl1, spl2 and chosen_sample
			 * 
			 * but only if sample lookup is successful, else variables remain null 
			 * 
			 * The method delegates to methods specific for when using INDEX_FILE or not  
			 */
			 
			String [] arr = sampleLookup( bc1, bc2, bc2sample);
			spl1 = arr[0];
			spl2 = arr[1];
			chosen_sample = arr[2];
			
			if(chosen_sample!=null){
				fwdW = writers_fwd.get(chosen_sample);
				if(RUNNING_PAIRED_END)
					revW = writers_rev.get(chosen_sample);
			}
			
			log.debug("spl1="+spl1);
			log.debug("spl2="+spl2);
			log.debug("chosen_sample="+chosen_sample);
			if(bc1!=null)
				log.debug("bc1="+bc1.toString());
			else
				log.debug("bc1=NULL");
			
			if(bc2!=null)
				log.debug("bc2="+bc2.toString());
			else
				log.debug("bc2=NULL");
			
			//if chosen_sample == null, the read pair is unassigned
			if(chosen_sample == null){
				unassigned++;
				if(KEEP_UNASSIGNED_READ){
					fwdW = writers_fwd.get(UNASSIGNED);
					if(RUNNING_PAIRED_END) 
						revW = writers_rev.get(UNASSIGNED);
				}
			}
			else{
				assigned++;
				//increase count for sample
				int c = sampleCountMap.get(chosen_sample);
				c++;
				sampleCountMap.put(chosen_sample, c);
			}

			//write diag no matter if read pair is assigned to a sample or not
			if(writingBarcodeDiagFile){
				String [] vals = null;
				if(RUNNING_PAIRED_END){
					//	headers = new String []{"Read_1_Header", "Read_2_Header", "seq1", "seq2" , "MM_Best_1", "MM_Second_1", "bc1", "smpl_1", "MM_Best_2", "MM_Second_2", "bc2", "smpl_2", "sample"};
					String mm1 = bc1 == null ? "" : bc1.mismatches+"";
					String mms1 = bc1 == null ? "" : bc1.mismatchesToSecondBest+"";
					String _bc1 =  bc1 == null ? "" : bc1.barcode;
					String mm2 = bc2 == null ? "" : bc2.mismatches+"";
					String mms2 = bc2 == null ? "" : bc2.mismatchesToSecondBest+"";
					String _bc2 =  bc2 == null ? "" : bc2.barcode;
					vals = new String []{
							r1.getReadHeader(), r2.getReadHeader(), s1, s2, 
							mm1 , mms1 , _bc1, spl1 , 
							mm2 , mms2 , _bc2, spl2 ,
							chosen_sample};
				}else{
					//headers =  new String []{"Read_Header", "seq", "MM_Best", "MM_Second", "bc", "sample"};
					String mm1 = bc1 == null ? "" : bc1.mismatches+"";
					String mms1 = bc1 == null ? "" : bc1.mismatchesToSecondBest+"";
					String _bc1 =  bc1 == null ? "" : bc1.barcode;

					vals = new String []{r1.getReadHeader(), s1, mm1 , mms1 , _bc1, chosen_sample};
				}

				diagPW.println(StringUtil.mergeArray(vals,"\t"));
			}

			if(!KEEP_UNASSIGNED_READ && chosen_sample == null){
				//then unassigned reads should not be saved, skip read writing
				continue;
			}

			/*
			 * write reads to out files
			 */
			if(!STATS_ONLY){
				if(USE_SAMPLE_INDEX_FILES){
					/*
					 * up to now, index files can only barcode encoding sample information ie no Ns can be found in there
					 * But there might be additional barcodes in the reads to extract
					 */
					
					if(RUNNING_PAIRED_END){
						String bcForRead1 = (bc1 == null ? null : bc1.barcode);
						String bcForRead2 = (bc2 == null ? null : bc2.barcode);
						if(INDEX_FILE2 == null ){
							//then it is the same barcode for read2 as read1
							bcForRead2= (bc1 == null ? null : bc1.barcode);
						}else if(!REDUNDANT_BARCODES){
							//when 2 INDEX_FILES and BRED=false => both barcodes should be added to each read like I1:I2
							String concatBCs =  bcForRead1+":"+bcForRead2;
							bcForRead1 = concatBCs;
							bcForRead2 = concatBCs;
						}
						write(r1, BCLEN_1, XTRIMLEN_1, ZTRIMLEN_1, fwdW, bcForRead1, ADD_BARCODE_TO_HEADER);
						write(r2, BCLEN_2 , XTRIMLEN_2 , ZTRIMLEN_2, revW, bcForRead2, ADD_BARCODE_TO_HEADER );
					}else{
						//single end
						write(r1, BCLEN_1, XTRIMLEN_1, ZTRIMLEN_1, fwdW, (bc1 == null ? null : bc1.barcode), ADD_BARCODE_TO_HEADER);
					}
					
				}
				else{

					/*
					 * special situation where barcodes have N's : give extracted sequence from read, not the matched barcodes
					 */
					if( barcodeValidator.barcodesHaveNs ){
						if(RUNNING_PAIRED_END){
							String bcForRead1 = (bc1 == null ? null : s1);
							String bcForRead2 = (bc2 == null ? null : s2);
							if(BARCODE_READ_POS == BarcodePosition.READ_2 ){
								bcForRead1 = (bc2 == null ? null : s2);
							}
							else if(BARCODE_READ_POS==BarcodePosition.READ_1){
								bcForRead2= (bc1 == null ? null : s1);
							}
							write(r1, BCLEN_1, XTRIMLEN_1, ZTRIMLEN_1, fwdW, bcForRead1, ADD_BARCODE_TO_HEADER);
							write(r2, BCLEN_2 , XTRIMLEN_2 , ZTRIMLEN_2, revW, bcForRead2, ADD_BARCODE_TO_HEADER );
						}else{
							//single end
							write(r1, BCLEN_1, XTRIMLEN_1, ZTRIMLEN_1, fwdW, (bc1 == null ? null : s1), ADD_BARCODE_TO_HEADER);
						}
					}else{
						//normal situation
						if(RUNNING_PAIRED_END){
							String bcForRead1 = (bc1 == null ? null : bc1.barcode);
							String bcForRead2 = (bc2 == null ? null : bc2.barcode);
							if(BARCODE_READ_POS == BarcodePosition.READ_2 ){
								bcForRead1 = (bc2 == null ? null : bc2.barcode);
							}
							else if(BARCODE_READ_POS==BarcodePosition.READ_1){
								bcForRead2= (bc1 == null ? null : bc1.barcode);
							}
							write(r1, BCLEN_1, XTRIMLEN_1, ZTRIMLEN_1, fwdW, bcForRead1, ADD_BARCODE_TO_HEADER);
							write(r2, BCLEN_2 , XTRIMLEN_2 , ZTRIMLEN_2, revW, bcForRead2, ADD_BARCODE_TO_HEADER );
						}else{
							//single end
							write(r1, BCLEN_1, XTRIMLEN_1, ZTRIMLEN_1, fwdW, (bc1 == null ? null : bc1.barcode), ADD_BARCODE_TO_HEADER);
						}
					}
				}
			}

		}

		//close all writers
		if(!STATS_ONLY){
			for (FastqWriter w : writers_fwd.values()) {
				if(w!=null) w.close(); 
			}
			if(RUNNING_PAIRED_END){
				for (FastqWriter w : writers_rev.values()) {
					if(w!=null) w.close();
				}
			}
		}
		if(writingBarcodeDiagFile){
			if(diagPW!=null) diagPW.close();
		}

		//print counts
		PrintWriter pw = null;
		try {
			pw = new PrintWriter(metricsFile);
			pw.println("Processed_read_pairs"+"\t"+ cnt);
			pw.println("Assigned"+"\t"+ assigned );
			pw.println("Unassigned_read_pairs"+"\t"+ unassigned);
			for (Entry<String, Integer> e : sampleCountMap.entrySet()) {
				pw.println(e.getKey()+"\t"+ e.getValue() );
			}

		} catch (Exception e) {
			log.error(ExceptionUtil.getStackTrace(e));
			throw new RuntimeException(e);
		} finally {
			if(pw!=null)
				pw.close();
		}

		//all fine

		return 0;

	}


	/**
	 * wrapper method to sampleLookupWithoutIndexFiles(..) and sampleLookupWithIndexFiles(..)
	 * 
	 * @param bc1
	 * @param bc2
	 * @param bc2sample
	 * @return a Strign array containing, in this order, the new values for 'spl1', 'spl2' and 'chosen_sample'
	 */
	private String[] sampleLookup(BarcodeMatch bc1, BarcodeMatch bc2, Map<String, String> bc2sample) {
		
		if(USE_SAMPLE_INDEX_FILES){
			return sampleLookupWithIndexFiles(bc1, bc2, bc2sample);
		}else{
			return sampleLookupWithoutIndexFiles( bc1, bc2, bc2sample);
		}
		
		
	}

	
	/**
	 * 
	 * sample matching in case of use of index files (I1 and I2)
	 * 
	 * @param bc1
	 * @param bc2
	 * @param bc2sample
	 * 
	 * @return a String array containing, in this order, the new values for 'spl1', 'spl2' and 'chosen_sample'
	 */
	private String [] sampleLookupWithIndexFiles(BarcodeMatch bc1, BarcodeMatch bc2, 
			Map<String, String> bc2sample){
		
		String spl1 = null;
		String spl2 = null;
		String chosen_sample = null;
		
		if(bc1!=null)
			log.debug("[sampleLookupWithIndexFiles] bc1="+bc1.toString());
		else
			log.debug("[sampleLookupWithIndexFiles] bc1=NULL");
		
		if(bc2!=null)
			log.debug("[sampleLookupWithIndexFiles] bc2="+bc2.toString());
		else
			log.debug("[sampleLookupWithIndexFiles] bc2=NULL");
		
		//only one barcode then it MUST be bc1 as INDEX_FILE2 == null
		if(INDEX_FILE2 == null){
			chosen_sample = (bc1==null || !bc1.matched) ? null : bc2sample.get(bc1.barcode); 
			return new String[]{chosen_sample, null, chosen_sample};
		}
		
		/*
		 * We then have 2 barcodes.
		 * Case 1 : both barcodes are redundant
		 */
		if(REDUNDANT_BARCODES == true){
			log.debug("[sampleLookupWithIndexFiles] 2 index file with redundant barcodes");
			spl1 = (bc1==null || !bc1.matched) ? null : bc2sample.get(bc1.barcode);
			spl2 = (bc2==null || !bc2.matched) ? null : bc2sample.get(bc2.barcode);

			if(spl1 != null &&  spl2 != null){
				if(spl1.equals(spl2)){
					chosen_sample=spl1;
				}else if(!STRICT){
					//which is the best match ?
					if(bc1.mismatches > bc2.mismatches){
						//bc2 is best
						chosen_sample=spl2;
					}
					else if(bc1.mismatches < bc2.mismatches){
						//bc1 is best
						chosen_sample=spl1;
					}
					else{
						//look int delta
						if(bc1.mismatchesToSecondBest!=bc2.mismatchesToSecondBest){
							chosen_sample = bc1.mismatchesToSecondBest>bc2.mismatchesToSecondBest? spl1:spl2;
						}
					}
				}
			}
			else if(!STRICT && spl1 != null){
				chosen_sample=spl1;
			}
			else if(!STRICT && spl2 != null){
				chosen_sample=spl2;					
			}
			
			return new String[]{spl1, spl2, chosen_sample};
		}
		/*
		 * Case 2: both barcodes are needed to lookup sample
		 */
		else {
			log.debug("[sampleLookupWithIndexFiles] 2 index file with non-redundant barcodes");
			//we need a match for both 
			if(bc1!=null && bc1.matched && bc2!=null && bc2.matched){
				spl1="NA";
				spl2="NA";
				//then the lookup is made with concatenated barcodes 
				String _bc = bc1.barcode+bc2.barcode;
				chosen_sample = bc2sample.get(_bc);
				
			}
			
			return new String[]{spl1, spl2, chosen_sample};
		}
		
		
	}
	
	/**
	 * 
	 * Sample matching in the normal case of not using index files ie barcode embedded in read(s)
	 * 
	 * @param bc1
	 * @param bc2
	 * @param bc2sample
	 * @return a String array containing, in this order, the new values for 'spl1', 'spl2' and 'chosen_sample'
	 */
	private String [] sampleLookupWithoutIndexFiles(BarcodeMatch bc1, BarcodeMatch bc2, Map<String, String> bc2sample){

		
		String spl1 = null;
		String spl2 = null;
		String chosen_sample = null;
		
		//if running SE and we could match up a barcode
		if(!RUNNING_PAIRED_END && bc1!=null &&bc1.matched){
			chosen_sample = bc2sample.get(bc1.barcode);
			
			//stop method
			return new String[]{chosen_sample, null, chosen_sample};
		}


		/*
		 * if only one barcode should be used for sample look up (note: no matter if barcodes are present at both ends)
		 */				
		if(BARCODE_FOR_SAMPLE_MATCHING != BarcodePosition.BOTH ){
			BarcodeMatch bm = BARCODE_FOR_SAMPLE_MATCHING==BarcodePosition.READ_1 ? bc1:bc2;
			String spl = (bm==null || !bm.matched) ? null : bc2sample.get(bm.barcode);

			if(spl!=null){
				chosen_sample = spl;
			}
			
			return new String[]{
					BARCODE_FOR_SAMPLE_MATCHING==BarcodePosition.READ_1 ? chosen_sample:null, 
							BARCODE_FOR_SAMPLE_MATCHING==BarcodePosition.READ_2 ? chosen_sample:null, 
									chosen_sample};
		}
		/*
		 * if both barcodes are redundant
		 */
		else if(BARCODE_FOR_SAMPLE_MATCHING == BarcodePosition.BOTH && REDUNDANT_BARCODES == true){
			spl1 = (bc1==null || !bc1.matched) ? null : bc2sample.get(bc1.barcode);
			spl2 = (bc2==null || !bc2.matched) ? null : bc2sample.get(bc2.barcode);

			if(spl1 != null &&  spl2 != null){
				if(spl1.equals(spl2)){
					chosen_sample=spl1;
				}else if(!STRICT){
					//which is the best match ?
					if(bc1.mismatches > bc2.mismatches){
						//bc2 is best
						chosen_sample=spl2;
					}
					else if(bc1.mismatches < bc2.mismatches){
						//bc1 is best
						chosen_sample=spl1;
					}
					else{
						//look int delta
						if(bc1.mismatchesToSecondBest!=bc2.mismatchesToSecondBest){
							chosen_sample = bc1.mismatchesToSecondBest>bc2.mismatchesToSecondBest? spl1:spl2;
						}
					}
				}
			}
			else if(!STRICT && spl1 != null){
				chosen_sample=spl1;
			}
			else if(!STRICT && spl2 != null){
				chosen_sample=spl2;					
			}

			
			return new String[]{spl1, spl2, chosen_sample};

		}
		/*
		 * if both barcode are needed to lookup sample
		 */
		else if(BARCODE_FOR_SAMPLE_MATCHING == BarcodePosition.BOTH && REDUNDANT_BARCODES == false){
			//we need a match for both 
			if(bc1!=null && bc1.matched && bc2!=null && bc2.matched){
				spl1="NA";
				spl2="NA";
				//then the lookup is made with concatenated barcodes 
				chosen_sample = bc2sample.get(bc1.barcode+bc2.barcode);
				
			}
			
			return new String[]{spl1, spl2, chosen_sample};
		}
		else{
			//unexpected situation reflecting a bug
			throw new RuntimeException("Don t know how to rematch barcode(s) to sample : unexpected code dead end reached ! Please report.");
		}


	}
	

	private boolean isAbsolutePath(String fname) {
		
		if(fname.contains("/")){
			File f = new File(fname);
			return f.getParentFile().exists() && f.getParentFile().isDirectory();
		}
		return false;
	}


	/**
	 * @param r the read to write
	 * @param bcLen the barcode length
	 * @param xtrim how many extra base should be clipped after the barcode ; 0 or more
	 * @param writer the writer
	 * @param barcode the barcode sequence extracted from this read. Note that the caller is responsible to give the 
	 *  sequence as extracted from the read or the barcode sequence after matching the extracted sequence against list of expected barcode
	 * @param addBarcodeToHeader whether the barcode should be added to the read header. If true, the string ':barcode' is added 
	 * with the starting ':' added only if current read header does not end with ':'. For example :
	 * '@D3FCO8P1:178:C1WLBACXX:7:1101:1836:1965 2:N:0:'
	 * becomes
	 * '@D3FCO8P1:178:C1WLBACXX:7:1101:1836:1965 2:N:0:BARCODE'
	 * @return the trimmed {@link FastqRecord} ; mostly for testing purposes
	 */
	protected FastqRecord write(final FastqRecord r, int bcLen, int xtrim, int ztrim, final FastqWriter writer, String barcode, boolean addBarcodeToHeader) {
		
		
		if(USE_SAMPLE_INDEX_FILES){
			return writeWithIndexFiles(r, bcLen, xtrim, ztrim, writer, barcode, addBarcodeToHeader); 
		}
		
		int l = bcLen + xtrim ;
		String header = r.getReadHeader();
		if(addBarcodeToHeader){
			header+= (header.endsWith(":")? "":":")+(barcode!=null?barcode:"");
			if(READ_NAME_REPLACE_CHAR != null){
				header = header.replaceAll(" ", READ_NAME_REPLACE_CHAR);
			}
		}


		String subseq_trimmed = r.getReadString();
		String qualstring_trimmed = r.getBaseQualityString();
		if(CLIP_BARCODE && ztrim > 0){
			subseq_trimmed = r.getReadString().substring(l, subseq_trimmed.length() - ztrim) ;
			qualstring_trimmed = r.getBaseQualityString().substring(l, qualstring_trimmed.length() - ztrim) ;
		}
		else if(CLIP_BARCODE){
			subseq_trimmed = r.getReadString().substring(l) ;
			qualstring_trimmed = r.getBaseQualityString().substring(l) ;
		}
		else if(ztrim > 0){
			subseq_trimmed = r.getReadString().substring(0, subseq_trimmed.length() - ztrim) ;
			qualstring_trimmed = r.getBaseQualityString().substring(0, qualstring_trimmed.length() - ztrim) ;
		}


		FastqRecord trimmed = new FastqRecord(
				header,//seqHeaderPrefix
				subseq_trimmed, //seqLine,
				r.getBaseQualityHeader(),//qualHeaderPrefix,
				qualstring_trimmed // qualLine
				);

		if(writer != null) //mostly here for testing ie to be able to call method w/o writer
			writer.write(trimmed);

		return trimmed;
	}

	/**
	 * Writes the FastqRecord in the case of using index files
	 * 
	 * @param r the read to write, as read from input FASTQ file
	 * @param bcLen the barcode length. In this case, this is the length of an extra barcode found at the read start
	 * @param xtrim how many extra base should be clipped after the barcode ; 0 or more
	 * @param writer the writer
	 * @param barcode the barcode sequence from this read. Note that the caller is responsible to give the 
	 *  sequence as extracted from the read or the barcode sequence after matching the extracted sequence against list of expected barcode
	 * @param addBarcodeToHeader whether the barcode should be added to the read header. If true, the string ':barcode' is added 
	 * with the starting ':' added only if current read header does not end with ':'. For example :
	 * '@D3FCO8P1:178:C1WLBACXX:7:1101:1836:1965 2:N:0:'
	 * becomes
	 * '@D3FCO8P1:178:C1WLBACXX:7:1101:1836:1965 2:N:0:BARCODE'
	 * @return the trimmed {@link FastqRecord} ; mostly for testing purposes
	 */
	private FastqRecord writeWithIndexFiles(FastqRecord r, int bcLen,
			int xtrim, int ztrim, FastqWriter writer, String barcode,
			boolean addBarcodeToHeader) {
		
		
		int l = bcLen + xtrim ; //safe for bcLen = 0
		log.debug(l);
		//create the new header
		String header = r.getReadHeader();
		if(addBarcodeToHeader){
			String extractedBCSeq = bcLen > 0 ? ":"+r.getReadString().substring(0, bcLen).toUpperCase() : "";
			
			header+= (header.endsWith(":")? "":":")+(barcode!=null?barcode:"")+extractedBCSeq;
			if(READ_NAME_REPLACE_CHAR != null){
				header = header.replaceAll(" ", READ_NAME_REPLACE_CHAR);
			}
		}

		String subseq_trimmed = r.getReadString();
		String qualstring_trimmed = r.getBaseQualityString();
		if(CLIP_BARCODE && ztrim > 0){
			subseq_trimmed = r.getReadString().substring(l, subseq_trimmed.length() - ztrim) ;
			qualstring_trimmed = r.getBaseQualityString().substring(l, qualstring_trimmed.length() - ztrim) ;
		}
		else if(CLIP_BARCODE){
			subseq_trimmed = r.getReadString().substring(l) ;
			qualstring_trimmed = r.getBaseQualityString().substring(l) ;
		}
		else if(ztrim > 0){
			subseq_trimmed = r.getReadString().substring(0, subseq_trimmed.length() - ztrim) ;
			qualstring_trimmed = r.getBaseQualityString().substring(0, qualstring_trimmed.length() - ztrim) ;
		}
		

		FastqRecord trimmed = new FastqRecord(
				header,//seqHeaderPrefix
				subseq_trimmed, //seqLine,
				r.getBaseQualityHeader(),//qualHeaderPrefix,
				qualstring_trimmed // qualLine
				);

		if(writer != null) //mostly here for testing ie to be able to call method w/o writer
			writer.write(trimmed);

		return trimmed;
	}


	/**
	 * 
	 * Finds best matching barcode according to global settings i.e. 
	 * MINIMUM_BASE_QUALITY
	 * MIN_MISMATCH_DELTA
	 * MAX_MISMATCHES
	 * 
	 * @param barcodePosition tell what barcode we are currently matching i.e. either READ_1 or READ_2 
	 * @param readSubsequence the read subsequence to match against barcodes
	 * @param qualitySubsequence the quality subsequence corresponding to s
	 * @param bc2sample
	 * @param sample2bcSet
	 * @return a {@link BarcodeMatch} describing the match or null if no match was found
	 */
	protected BarcodeMatch findBarcode(
			final BarcodePosition barcodePosition,
			final String readSubsequence,
			final String qualitySubsequence,
			final Map<String, String> bc2sample,
			final Map<String, Set<String>> sample2bcSet) {

		if(readSubsequence == null)
			return null;
		
		if(barcodePosition == BarcodePosition.BOTH)
			throw new RuntimeException("findBarcode method can t be called with barcodePosition = BOTH");

		//fecth set of allowed barcodes at this positin
		Set<String> possibleBarcodes = barcodePosition == BarcodePosition.READ_1 ? barcodeValidator.barcodeSetRead1 : barcodeValidator.barcodeSetRead2;
		
		boolean barcodesHaveNs = barcodeValidator.areNsInBarcodes(barcodePosition);
		/*
		 * we first check if the subsequence matches a known barcode, in which case, we directly take this as a result, 
		 * BUT this cannot work if we have Ns in barcodes (like in iCLIP), to speed up code we directly use a predefined 
		 */
		if(!barcodesHaveNs && possibleBarcodes.contains(readSubsequence)){
			BarcodeMatch bm = new BarcodeMatch();
			bm.matched = true;
			bm.barcode = readSubsequence;
			bm.mismatches = 0;
			bm.mismatchesToSecondBest = readSubsequence.length();
			return bm;
		}else{
			BarcodeMatch bm =  findBestBarcode(barcodePosition, readSubsequence, qualitySubsequence); 
			if(bm.matched)
				return bm;

		}
		return null;
	}



	/**
	 * Find the best barcode match for the given read sequence
	 * @param barcodePosition the {@link BarcodePosition} 
	 * @param readSubsequence portion of read containing barcode
	 * @param qualityScoreString the string representing the quality scores for the read subsequence
	 * @return a BarcodeMatch holding the matched status and extra info when matched is true
	 */
	protected BarcodeMatch findBestBarcode(final BarcodePosition barcodePosition, final String readSubsequence, final String qualityScoreString) {

		//turn to bytes
		byte [] subseqBytes = net.sf.samtools.util.StringUtil.stringToBytes(readSubsequence);
		byte[] qualBytes = convertQualityStringToByteArray(qualityScoreString);

		int numMismatchesInBestBarcode = readSubsequence.length() + 1; //init with max mismatch num + 1
		int numMismatchesInSecondBestBarcode = readSubsequence.length() + 1;
		//find best and second best barcode
		byte[] bestBarcode = null;
		
		//look up adequate variables
		byte [][] barcodeBytes = barcodePosition == BarcodePosition.READ_1 ? barcodeBytes1 : barcodeBytes2;
		int minQuality = barcodePosition == BarcodePosition.READ_1 ? MIN_BASE_QUALITY_1 : MIN_BASE_QUALITY_2;
		int maxMismatches = barcodePosition == BarcodePosition.READ_1 ? MAX_MISMATCHES_1 : MAX_MISMATCHES_2;
		int minMMDelta = barcodePosition == BarcodePosition.READ_1 ? MIN_MISMATCH_DELTA_1 : MIN_MISMATCH_DELTA_2;
		
		for (int i = 0; i < barcodeBytes.length; i++) {
			final int numMismatches = countMismatches(barcodeBytes[i], subseqBytes, qualBytes, minQuality);
			if (numMismatches < numMismatchesInBestBarcode) {
				if (bestBarcode != null) {
					numMismatchesInSecondBestBarcode = numMismatchesInBestBarcode;
				}
				numMismatchesInBestBarcode = numMismatches;
				bestBarcode = barcodeBytes[i];
			} else if (numMismatches < numMismatchesInSecondBestBarcode) {
				numMismatchesInSecondBestBarcode = numMismatches;
			}
		}


		final boolean matched = (bestBarcode != null &&
				numMismatchesInBestBarcode <= maxMismatches &&
				numMismatchesInSecondBestBarcode - numMismatchesInBestBarcode >= minMMDelta)
				;


		final BarcodeMatch match = new BarcodeMatch();
		match.matched = matched;

		if (matched) {
			match.mismatches = numMismatchesInBestBarcode;
			match.mismatchesToSecondBest = numMismatchesInSecondBestBarcode;
			match.barcode = net.sf.samtools.util.StringUtil.bytesToString( bestBarcode );
		}


		return match;
	}



	/**
	 * This considers a byte offset of 33 i.e. Phred+33 (Illumina compatible)
	 * Method inspired from {@link FastqToSam}
	 * @param qualityScores the quality scire string from Illumina
	 * @return
	 */
	protected byte[] convertQualityStringToByteArray(final String qualityScores) {
		byte [] qualBytes = net.sf.samtools.util.StringUtil.stringToBytes(qualityScores);
		convertQuality(qualBytes, QUALITY_FORMAT);
		return qualBytes;
	}

	/** 
	 * Based on the type of quality scores coming in, converts them to a numeric byte[] in phred scale. 
	 */
	void convertQuality(final byte[] quals, final FastqQualityFormat version) {
		switch (version)  {
		case Standard:
			SAMUtils.fastqToPhred(quals);
			break ;
		case Solexa:
			solexaQualityConverter.convertSolexaQualityCharsToPhredBinary(quals);
			break ;
		case Illumina:
			solexaQualityConverter.convertSolexa_1_3_QualityCharsToPhredBinary(quals);
			break ;
		}
	}



	/**
	 * 
	 * Compare barcode sequence to bases from read
	 * @param barcodeBytes the barcode as array of bytes
	 * @param readSubsequence the subsequence as array of bytes
	 * @param qualities the base quality score in phred scale (+33)
	 * @param minimumBaseQuality an int in [0,40]
	 * @return how many bases did not match , 'N' or base with quality less than MINIMUM_BASE_QUALITY always considered
	 * as a mismatch 
	 * 
	 */
	protected  int countMismatches(final byte[] barcodeBytes, final byte[] readSubsequence, final byte[] qualities, final int minimumBaseQuality) {
		int numMismatches = 0;
		// Read sequence and barcode length may not be equal, so we just use the shorter of the two ; I don t see why this should happen but it is free to do this !
		final int basesToCheck = Math.min(barcodeBytes.length, readSubsequence.length);
		for (int i = 0; i < basesToCheck; ++i) {
			/*
			 * we first need to check this is a valid position in barcode (barcode can contain N to specify this position should be ignored)
			 * If the bc contains a N, we skip
			 */
			if (!net.sf.samtools.util.SequenceUtil.isValidBase(barcodeBytes[i])) { // Returns true if the byte is in [acgtACGT].
				continue;
			}
			
			if (!net.sf.samtools.util.SequenceUtil.isNoCall(readSubsequence[i])) { //returns true if the value of base represents a no call
				//if bases in barcode and readsubsequence are different => increase mismatch count ; else check quality
				if (!net.sf.samtools.util.SequenceUtil.basesEqual(barcodeBytes[i], readSubsequence[i])){
					++numMismatches;
				}
				else if (qualities != null){
					final int uQual = qualities[i] & 0xff; //from FastqToSam.java ; this is kind of bit masking 
					if(uQual < minimumBaseQuality)
						++numMismatches;
				}
			}else{
				//the read subsequence contains a 'N' at this position => this is a mismatch 
				++numMismatches;
			}
		}

		return numMismatches;
	}


	/**
	 * @return the result of executing whoami on the underlying OS 
	 */
	protected String whoami() {
		String command = "whoami" ;
        Process myprocess = null;
        try {
            myprocess = Runtime.getRuntime().exec(command);
            ProcessConsumer pc = new ProcessConsumer(myprocess);
            StringBuffer err = new StringBuffer();
            StringBuffer out = new StringBuffer();
            
            int rc = pc.output( out ) 
                .error( err ) 
                .consume();
            
            if (rc != 0) {
                return null;
            }
            return	out.toString().replaceAll("\n", "");
            
            
        } catch (Exception e) {
			return null;
		}finally{
            if(myprocess!=null){
            	myprocess.destroy();
				myprocess = null;
            }
        }
         
		
	}
	
	/**
	 * @param args
	 */
	public static void main(final String[] argv) {
		new Jemultiplexer().instanceMainWithExit(argv); //eventually execute the doWork() method
	}

	
	

}
